import numpy as np
import scipy.sparse as sp

from ..simulation import BaseSimulation
from ..survey import BaseSurvey
from ..maps import IdentityMap
from ..utils import validate_list_of_types, validate_type
from ..props import HasModel
import itertools
import warnings


class MetaSimulation(BaseSimulation):
    """Combine multiple simulations into a single one.

    This class is used to encapsulate multiple simulations into a
    single simulation. Each simulation and mapping pair will
    perform its own work, then concatenate the results together.

    For each mapping and simulation pair, given a model, this first
    applies the mapping, then passes the resulting model to the simulation.

    With the proper mappings this can be useful for setting up time-lapse,
    tiled, stitched, or any other simulation that can be broken into many
    individual simulations.

    Parameters
    ----------
    simulations : (n_sim) list of SimPEG.simulation.BaseSimulation
        The list of unique simulations that each handle a piece
        of the problem.
    mappings : (n_sim) list of SimPEG.maps.IdentityMap
        The map for every simulation. Every map should accept the
        same length model, and output a model appropriate for its
        paired simulation.

    Examples
    --------
    Create a list of 1D simulations that perform a piece of a
    stitched problem.

    >>> from SimPEG.simulation import ExponentialSinusoidSimulation
    >>> from SimPEG import maps
    >>> from SimPEG.meta import MetaSimulation
    >>> from discretize import TensorMesh
    >>> import matplotlib.pyplot as plt

    Create a mesh for space and time, then one that represents
    the full dimensionality of the model.
    >>> mesh_space = TensorMesh([100])
    >>> mesh_time = TensorMesh([5])
    >>> full_mesh = TensorMesh([5, 100])

    Lets say we have observations at 5 locations in time. For simplicity
    we will just use the same times from the time mesh, but this is not
    required. Then create a simulation for each of these times. We also
    create an operator that maps the model in full space to the model for
    each simulation.
    >>> obs_times = mesh_time.cell_centers_x
    >>> sims, mappings = [], []
    >>> for time in obs_times:
    ...     sims.append(ExponentialSinusoidSimulation(
    ...         mesh=mesh_space,
    ...         model_map=maps.IdentityMap(),
    ...     ))
    ...     ccs = mesh_space.cell_centers
    ...     p_ave = full_mesh.get_interpolation_matrix(
    ...         np.c_[np.full_like(ccs, time), ccs]
    ...     )
    ...     mappings.append(maps.LinearMap(p_ave))
    >>> sim = MetaSimulation(sims, mappings)

    This simulation acts like a single simulation, which can be used for modeling
    and inversion. This model is a moving box car.
    >>> true_model = np.zeros(full_mesh.shape_cells)
    >>> speed, start, width = 0.8, 0.1, 0.2
    >>> for i, time in enumerate(mesh_time.cell_centers):
    ...     center = speed * time  + start
    ...     in_box = np.abs(mesh_space.cell_centers - center) <= width/2
    ...     true_model[i, in_box] = 1.0
    >>> true_model = true_model.reshape(-1, order='F')

    Then use the simulation to create data.
    >>> d_pred = sim.dpred(true_model)
    >>> plt.plot(d_pred.reshape(5, -1).T)
    >>> plt.show()
    """

    _repeat_sim = False

    def __init__(self, simulations, mappings):
        warnings.warn(
            "The MetaSimulation class is a work in progress and might change in the future",
            stacklevel=2,
        )
        self.simulations = simulations
        self.mappings = mappings
        self.model = None
        # give myself a BaseSurvey that has the number of data equal
        # to the sum of the sims' data.
        survey = BaseSurvey([])
        vnD = [sim.survey.nD for sim in self.simulations]
        survey._vnD = vnD
        self.survey = survey
        self._data_offsets = np.cumsum(np.r_[0, vnD])

    @property
    def simulations(self):
        """The list of simulations.

        Returns
        -------
        (n_sim) list of SimPEG.simulation.BaseSimulation
        """
        return self._simulations

    @simulations.setter
    def simulations(self, value):
        self._simulations = validate_list_of_types(
            "simulations", value, BaseSimulation, ensure_unique=True
        )

    @property
    def mappings(self):
        """The mappings paired to each simulation.

        Every mapping should accept the same length model, and output
        a model that is consistent with the simulation.

        Returns
        -------
        (n_sim) list of SimPEG.maps.IdentityMap
        """
        return self._mappings

    @mappings.setter
    def mappings(self, value):
        value = validate_list_of_types("mappings", value, IdentityMap)
        if not self._repeat_sim and len(value) != len(self.simulations):
            raise ValueError(
                "Must provide the same number of mappings and simulations."
            )
        model_len = value[0].shape[1]
        for i, (mapping, sim) in enumerate(zip(value, self.simulations)):
            if mapping.shape[1] != model_len:
                raise ValueError("All mappings must have the same input length")
            map_out_shape = mapping.shape[0]
            for name in sim._act_map_names:
                sim_mapping = getattr(sim, name)
                sim_in_shape = sim_mapping.shape[1]
                if (
                    map_out_shape != "*"
                    and sim_in_shape != "*"
                    and sim_in_shape != map_out_shape
                ):
                    raise ValueError(
                        f"Simulation and mapping at index {i} inconsistent. "
                        f"Simulation mapping shape {sim_in_shape} incompatible with "
                        f"input mapping shape {map_out_shape}."
                    )
        self._mappings = value

    @property
    def _act_map_names(self):
        # Implement this here to trick the model setter to know about
        # how long an input model should be.
        # essentially it points to the first mapping.
        return ["_model_map"]

    @property
    def _model_map(self):
        # all of the mappings have the same input shape, so just return
        # the first one.
        return self.mappings[0]

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, value):
        updated = HasModel.model.fset(self, value)
        # Only send the model to the internal simulations if it was updated.
        if not self._repeat_sim and updated:
            for mapping, sim in zip(self.mappings, self.simulations):
                if value is not None:
                    sim.model = mapping * self._model
                else:
                    sim.model = value

    def fields(self, m):
        """Create fields for every simulation.

        The returned list contains the field object from each simulation.

        Returns
        -------
        (n_sim) list
            The type of each item of the list is determined by the internal
            simulation that created it.
        """
        self.model = m
        # The above should pass the model to all the internal simulations.
        f = []
        for mapping, sim in zip(self.mappings, self.simulations):
            if self._repeat_sim and self.model is not None:
                sim.model = mapping * self.model
            f.append(sim.fields(sim.model))
        return f

    def dpred(self, m=None, f=None):
        if f is None:
            if m is None:
                m = self.model
            f = self.fields(m)
        d_pred = []
        for mapping, sim, field in zip(self.mappings, self.simulations, f):
            if self._repeat_sim:
                sim.model = mapping * self.model
            d_pred.append(sim.dpred(m=sim.model, f=field))
        return np.concatenate(d_pred)

    def Jvec(self, m, v, f=None):
        self.model = m
        if f is None:
            f = self.fields(m)
        j_vec = []
        for mapping, sim, field in zip(self.mappings, self.simulations, f):
            if self._repeat_sim:
                sim.model = mapping * self.model
            sim_v = mapping.deriv(self.model) @ v
            j_vec.append(sim.Jvec(sim.model, sim_v, f=field))
        return np.concatenate(j_vec)

    def Jtvec(self, m, v, f=None):
        self.model = m
        if f is None:
            f = self.fields(m)
        jt_vec = 0
        for i, (mapping, sim, field) in enumerate(
            zip(self.mappings, self.simulations, f)
        ):
            if self._repeat_sim:
                sim.model = mapping * self.model
            sim_v = v[self._data_offsets[i] : self._data_offsets[i + 1]]
            jt_vec += mapping.deriv(self.model).T @ sim.Jtvec(sim.model, sim_v, f=field)
        return jt_vec

    def getJtJdiag(self, m, W=None, f=None):
        """Return the squared sum of columns of the Jacobian.

        Evaluates the weighted squared norm of each column
        of the Jacobian matrix. This is usually used to construct
        sensitivity weighting matrices or for diagonal preconditioners
        to iterative solvers.

        Parameters
        ----------
        m : (n_m) numpy.ndarray
            The model to evalute the Jacobian at.
        W : (n_d, n_d) scipy.sparse.csr_matrix, optional
            A diagonal data weighting matrix.
        f : fields, optional
            The fields object created from this class.

        Returns
        -------
        (n_m) numpy.ndarray
            Squared sum of columns of the Jacobian matrix

        Notes
        -----
        Internally, this function evaluates the ``getJtJdiag`` method of each
        simulation, then applies the model mapping to the output as:

        >>> sq_sum = 0
        >>> for i in range(n_sim):
        ...    row = sim[i].getJtJdiag(model)
        ...    sq_sum += W[i] * (sp.diag(sqrt(row)) @ mapping[i].deriv()).power(2).sum(axis=0)

        This approach is correct for mapping that match input parameters to a
        single output parameter, (i.e. the `mapping.deriv` has only 1 element in each column).
        For other mappings, it is usually close within a scaling factor, whose accuracy is
        then controlled by how diagonally dominant ``J.T @ J`` is.
        """
        self.model = m
        if getattr(self, "_jtjdiag", None) is None:
            if W is None:
                W = np.ones(self.survey.nD)
            else:
                W = W.diagonal()
            jtj_diag = 0.0
            # approximate the JtJ diag on the full model space as:
            # sum((diag(sqrt(jtj_diag)) @ M_deriv))**2)
            # Which is correct for mappings that match input parameters to only 1 output parameter.
            # (i.e. projections, multipliers, etc.).
            # It is usually close within a scaling factor for others, whose accuracy is controlled
            # by how diagonally dominant JtJ is.
            if f is None:
                f = self.fields(m)
            for i, (mapping, sim, field) in enumerate(
                zip(self.mappings, self.simulations, f)
            ):
                if self._repeat_sim:
                    sim.model = mapping * self.model
                sim_w = sp.diags(W[self._data_offsets[i] : self._data_offsets[i + 1]])
                sim_jtj = sp.diags(np.sqrt(sim.getJtJdiag(sim.model, sim_w, f=field)))
                m_deriv = mapping.deriv(self.model)
                jtj_diag += np.asarray(
                    (sim_jtj @ m_deriv).power(2).sum(axis=0)
                ).flatten()
            self._jtjdiag = jtj_diag

        return self._jtjdiag

    @property
    def deleteTheseOnModelUpdate(self):
        return super().deleteTheseOnModelUpdate + ["_jtjdiag"]


class SumMetaSimulation(MetaSimulation):
    """An extension of the MetaSimulation that sums the data outputs.

    This class requires the mappings have the same input length
    and each simulation to have the same number of data.

    This could be useful for a linear problem where each simulation
    tackles a different subset of the model.

    Parameters
    ----------
    simulations : (n_sim) list of SimPEG.simulation.BaseSimulation
    mappings : (n_sim) list of SimPEG.maps.IdentityMap
    """

    _repeat_sim = False

    def __init__(self, simulations, mappings):
        warnings.warn(
            "The SumMetaSimulation class is a work in progress and might change in the future",
            stacklevel=2,
        )
        self.simulations = simulations
        self.mappings = mappings
        self.model = None
        # give myself a BaseSurvey
        survey = BaseSurvey([])
        survey._vnD = [
            self.simulations[0].survey.nD,
        ]
        self.survey = survey

    @MetaSimulation.simulations.setter
    def simulations(self, value):
        value = validate_list_of_types(
            "simulations", value, BaseSimulation, ensure_unique=True
        )
        nD = value[0].survey.nD
        for sim in value:
            if sim.survey.nD != nD:
                raise ValueError("All simulations must have the same number of data.")
        self._simulations = value

    def dpred(self, m=None, f=None):
        if f is None:
            if m is None:
                m = self.model
            f = self.fields(m)
        d_pred = 0
        for sim, field in zip(self.simulations, f):
            d_pred += sim.dpred(m=sim.model, f=field)
        return d_pred

    def Jvec(self, m, v, f=None):
        self.model = m
        if f is None:
            f = self.fields(m)
        j_vec = 0
        for mapping, sim, field in zip(self.mappings, self.simulations, f):
            # Every d_pred needs to be setup to grab the current model
            # if given m=None as an argument.
            sim_v = mapping.deriv(self.model) @ v
            j_vec += sim.Jvec(sim.model, sim_v, f=field)
        return j_vec

    def Jtvec(self, m, v, f=None):
        self.model = m
        if f is None:
            f = self.fields(m)
        jt_vec = 0
        for mapping, sim, field in zip(self.mappings, self.simulations, f):
            jt_vec += mapping.deriv(self.model).T @ sim.Jtvec(sim.model, v, f=field)
        return jt_vec

    def getJtJdiag(self, m, W=None, f=None):
        self.model = m
        if getattr(self, "_jtjdiag", None) is None:
            jtj_diag = 0.0
            if f is None:
                f = self.fields(m)
            for mapping, sim, field in zip(self.mappings, self.simulations, f):
                sim_jtj = sp.diags(np.sqrt(sim.getJtJdiag(sim.model, W, f=field)))
                m_deriv = mapping.deriv(self.model)
                jtj_diag += np.asarray(
                    (sim_jtj @ m_deriv).power(2).sum(axis=0)
                ).flatten()
            self._jtjdiag = jtj_diag

        return self._jtjdiag


class RepeatedSimulation(MetaSimulation):
    """A MetaSimulation where a single simulation is used repeatedly.

    This is most useful for linear simulations where a sensitivity matrix can be
    reused with different models. For non-linear simulations it will often be quicker
    to use the MetaSimulation class with multiple copies of the same simulation.

    Parameters
    ----------
    simulation : SimPEG.simulation.BaseSimulation
    mappings : (n_sim) list of SimPEG.maps.IdentityMap
    """

    _repeat_sim = True

    def __init__(self, simulation, mappings):
        warnings.warn(
            "The RepeatedSimulation class is a work in progress and might change in the future",
            stacklevel=2,
        )
        self.simulation = simulation
        self.mappings = mappings
        self.model = None
        survey = BaseSurvey([])
        vnD = len(self.mappings) * [self.simulation.survey.nD]
        survey._vnD = vnD
        self.survey = survey
        self._data_offsets = np.cumsum(np.r_[0, vnD])

    @property
    def simulations(self):
        return itertools.repeat(self.simulation)

    @property
    def simulation(self):
        """The internal simulation.

        Returns
        -------
        SimPEG.simulation.BaseSimulation
        """
        return self._simulation

    @simulation.setter
    def simulation(self, value):
        self._simulation = validate_type(
            "simulation", value, BaseSimulation, cast=False
        )



import numpy as np
from scipy.interpolate import LinearNDInterpolator, NearestNDInterpolator
import matplotlib.pyplot as plt
from matplotlib import colors
import warnings


def plot2Ddata(
    xyz,
    data,
    vec=False,
    nx=100,
    ny=100,
    ax=None,
    mask=None,
    level=False,
    figname=None,
    ncontour=10,
    dataloc=False,
    contourOpts=None,
    levelOpts=None,
    streamplotOpts=None,
    scale="linear",
    clim=None,
    method="linear",
    shade=False,
    shade_ncontour=100,
    shade_azimuth=-45.0,
    shade_angle_altitude=45.0,
    shadeOpts=None,
):
    """Interpolate and plot unstructured 2D data.

    General plotting for scalar and vector quantities as a function of their
    *x* and *y* locations. **plot2Ddata** uses interpolates the unstructured
    data to a specified set of gridded locations before plotting with
    :meth:`matplotlib.pyplot.contourf`. For vectors, :meth:`matplotlib.pyplot.streamplot`
    is used to add a stream plot. As this function produces a plot for
    2D data, the vertical position and vertical vector component (in the case
    of a vector) is ignored.

    Parameters
    ----------
    xyz : numpy.ndarray
        Data locations [x,y(,z)]. If the data locations are defined in 3D, the z-column is ignored.
    data : numpy.ndarray
        Data values. For scalar quantities, the data are stored in a 1D ``numpy.ndarray``.
        For vector quantities, data are stored in a numpy array of shape (N, dim).
    vec : bool
        If ``True``, the data values represent a vector quantity and the function
        creates a stream plot illustrating the *x* and *y* components of the vector.
    nx : int
        Number of grid locations along x-direction
    ny : int
        Number of grid locations along y-direction
    ax : matplotlib.axes
        An axes object on which to plot. If ``None``, the function creates an axes object
    mask : numpy.ndarray of bool
        Locations in the unstructured grid whose data are masked.
    level : boolean
        If ``True``, adds contours according to :meth:`matplotlib.pyplot.contour`
    figname : str
        Figure name
    ncontour : int
        number of contours in the contour plot
    dataloc : bool
        If ``True``, plot the data locations
    contourOpts : dict
        Dictionary defining keyword arguments when :meth:`matplotlib.pyplot.contourf` is called
    levelOpts : dict
        Dictionary defining keyword arguments when :meth:`matplotlib.pyplot.contourf` is called.
        This is only necessary when *level* = ``True``.
    clim : (2) numpy.ndarray
        Colorbar limits
    method : str
        Interpolation method used to approximate at gridded locations. Must be 'linear' or 'nearest'
    shade : bool
        If ``True``, add shading to the plot
    shade_ncontour : int
        Number of :meth:`matplotlib.pyplot.contourf` contours for the shading
    shade_azimuth : float
        Azimuthal angle for the light source if shading
    shade_angle_altitude : float
        Altitude angle for the light source if shading

    Returns
    -------
    cont : matplotlib.contour.ContourSet
        The filled contour plot
    ax : matplotlib.axes
        The axes object for the plot.
    CS : matplotlib.contour.ContourSet
        If the input parameter *levels* is ``True``, the function outputs
        the level set for the contours

    """

    # Error checking and set vmin, vmax
    vlimits = [None, None]

    if clim is not None:
        vlimits = [np.min(clim), np.max(clim)]

    if contourOpts is None:
        contourOpts = {}
    if levelOpts is None:
        levelOpts = {}
    if streamplotOpts is None:
        streamplotOpts = {}
    if shadeOpts is None:
        shadeOpts = {}

    for i, key in enumerate(["vmin", "vmax"]):
        if key in contourOpts.keys():
            if vlimits[i] is None:
                vlimits[i] = contourOpts.pop(key)
            else:
                if not np.isclose(contourOpts[key], vlimits[i]):
                    raise Exception(
                        "The values provided in the colorbar limit, clim {} "
                        "does not match the value of {} provided in the "
                        "contourOpts: {}. Only one value should be provided or "
                        "the two values must be equal.".format(
                            vlimits[i], key, contourOpts[key]
                        )
                    )
                contourOpts.pop(key)
    vmin, vmax = vlimits[0], vlimits[1]

    # create a figure if it doesn't exist
    if ax is None:
        fig = plt.figure()
        ax = plt.subplot(111)

    # interpolate data to grid locations
    xmin, xmax = xyz[:, 0].min(), xyz[:, 0].max()
    ymin, ymax = xyz[:, 1].min(), xyz[:, 1].max()
    x = np.linspace(xmin, xmax, nx)
    y = np.linspace(ymin, ymax, ny)
    X, Y = np.meshgrid(x, y)
    xy = np.c_[X.flatten(), Y.flatten()]

    if vec is False:
        if method == "nearest":
            F = NearestNDInterpolator(xyz[:, :2], data)
        else:
            F = LinearNDInterpolator(xyz[:, :2], data)
        DATA = F(xy)
        DATA = DATA.reshape(X.shape)

        # Levels definitions
        dataselection = np.logical_and(~np.isnan(DATA), np.abs(DATA) != np.inf)
        if scale == "log":
            DATA = np.abs(DATA)

        # set vmin, vmax if they are not already set
        vmin = DATA[dataselection].min() if vmin is None else vmin
        vmax = DATA[dataselection].max() if vmax is None else vmax

        if scale == "log":
            levels = np.logspace(np.log10(vmin), np.log10(vmax), ncontour + 1)
            norm = colors.LogNorm(vmin=vmin, vmax=vmax)
        else:
            levels = np.linspace(vmin, vmax, ncontour + 1)
            norm = colors.Normalize(vmin=vmin, vmax=vmax)

        if mask is not None:
            Fmask = NearestNDInterpolator(xyz[:, :2], mask)
            MASK = Fmask(xy)
            MASK = MASK.reshape(X.shape)
            DATA = np.ma.masked_array(DATA, mask=MASK)

        contourOpts = {"levels": levels, "norm": norm, "zorder": 1, **contourOpts}
        cont = ax.contourf(X, Y, DATA, **contourOpts)

        if level:
            levelOpts = {"levels": levels, "zorder": 3, **levelOpts}
            CS = ax.contour(X, Y, DATA, **levelOpts)

    else:
        # Assume size of data is (N,2)
        datax = data[:, 0]
        datay = data[:, 1]
        if method == "nearest":
            Fx = NearestNDInterpolator(xyz[:, :2], datax)
            Fy = NearestNDInterpolator(xyz[:, :2], datay)
        else:
            Fx = LinearNDInterpolator(xyz[:, :2], datax)
            Fy = LinearNDInterpolator(xyz[:, :2], datay)
        DATAx = Fx(xy)
        DATAy = Fy(xy)
        DATA = np.sqrt(DATAx**2 + DATAy**2).reshape(X.shape)
        DATAx = DATAx.reshape(X.shape)
        DATAy = DATAy.reshape(X.shape)
        if scale == "log":
            DATA = np.abs(DATA)

        # Levels definitions
        dataselection = np.logical_and(~np.isnan(DATA), np.abs(DATA) != np.inf)

        # set vmin, vmax
        vmin = DATA[dataselection].min() if vmin is None else vmin
        vmax = DATA[dataselection].max() if vmax is None else vmax

        if scale == "log":
            levels = np.logspace(np.log10(vmin), np.log10(vmax), ncontour + 1)
            norm = colors.LogNorm(vmin=vmin, vmax=vmax)
        else:
            levels = np.linspace(vmin, vmax, ncontour + 1)
            norm = colors.Normalize(vmin=vmin, vmax=vmax)

        if mask is not None:
            Fmask = NearestNDInterpolator(xyz[:, :2], mask)
            MASK = Fmask(xy)
            MASK = MASK.reshape(X.shape)
            DATA = np.ma.masked_array(DATA, mask=MASK)

        contourOpts = {"levels": levels, "norm": norm, "zorder": 1, **contourOpts}
        cont = ax.contourf(X, Y, DATA, **contourOpts)

        streamplotOpts = {"zorder": 4, "color": "w", **streamplotOpts}
        ax.streamplot(X, Y, DATAx, DATAy, **streamplotOpts)

        if level:
            levelOpts = {"levels": levels, "zorder": 3, **levelOpts}
            CS = ax.contour(X, Y, DATA, levels=levels, zorder=3, **levelOpts)

    if shade:

        def hillshade(array, azimuth, angle_altitude):
            """
            coded copied from https://www.neonscience.org/create-hillshade-py
            """
            azimuth = 360.0 - azimuth
            x, y = np.gradient(array)
            slope = np.pi / 2.0 - np.arctan(np.sqrt(x * x + y * y))
            aspect = np.arctan2(-x, y)
            azimuthrad = azimuth * np.pi / 180.0
            altituderad = angle_altitude * np.pi / 180.0
            shaded = np.sin(altituderad) * np.sin(slope) + np.cos(altituderad) * np.cos(
                slope
            ) * np.cos((azimuthrad - np.pi / 2.0) - aspect)
            return 255 * (shaded + 1) / 2

        shadeOpts = {
            "cmap": "Greys",
            "alpha": 0.35,
            "antialiased": True,
            "zorder": 2,
            **shadeOpts,
        }

        ax.contourf(
            X,
            Y,
            hillshade(DATA, shade_azimuth, shade_angle_altitude),
            shade_ncontour,
            **shadeOpts
        )

    if dataloc:
        ax.plot(xyz[:, 0], xyz[:, 1], "k.", ms=2)
    ax.set_aspect("equal", adjustable="box")
    if figname:
        plt.axis("off")
        fig.savefig(figname, dpi=200)
    if level:
        return cont, ax, CS
    else:
        return cont, ax


def plot_1d_layer_model(
    thicknesses,
    values,
    z0=0,
    scale="log",
    ax=None,
    plot_elevation=False,
    show_layers=False,
    vlim=None,
    **kwargs
):
    """
    Plot the vertical profile for a 1D layered Earth model.

    Parameters
    ----------
    thicknesses : list or numpy.ndarray of float
        A ``list`` or ``numpy.ndarray`` containing the layer thicknesses from the top layer down
    values : list or numpy.ndarray of float
        A ``list`` or ``numpy.ndarray`` containing the physical property values from the top layer down
    z0 : float
        Elevation of the surface
    scale: {'linear', 'log'}
        Plot physical property values on a linear or log10 scale.
    ax: matplotlib.axes.Axes, optional
        An axis object for the plot
    plot_elevation : bool
        If ``False``, the yaxis will be the depth. If ``True``, the yaxis is the elevation.
    show_layers : bool
        Plot horizontal lines to denote layers.
    vlim : tuple, optional
        The limits for the x-axis.

    Returns
    -------
    matplotlib.axes.Axes
        The axis object that holds the plot

    """

    if len(thicknesses) < len(values):
        thicknesses = np.r_[thicknesses, thicknesses[-1]]
    z_grid = np.r_[0.0, np.cumsum(thicknesses)]
    resistivity = np.repeat(values, 2)
    if vlim is None:
        v_min = 0.9 * np.min(values)
        v_max = 1.1 * np.max(values)
    else:
        v_min, v_max = vlim

    z = []
    for i in range(0, len(thicknesses)):
        z.append(np.r_[z_grid[i], z_grid[i + 1]])
    z = np.hstack(z)
    z[-1] = 1e200  # A really large number

    if plot_elevation:
        y_label = "Elevation (m)"
        z = z0 - z
        z_grid = z0 - z_grid
        flip_axis = False
    else:
        y_label = "Depth (m)"
        flip_axis = True

    if ax is None:
        fig = plt.figure(figsize=(6, 6))
        ax = fig.add_axes([0.15, 0.15, 0.75, 0.75])

    if show_layers:
        for locz in z_grid:
            plt.plot(
                np.linspace(v_min, v_max, 2),
                np.ones(2) * locz,
                "k--",
                lw=0.5,
                label="_nolegend_",
            )

    ax.plot(resistivity, z, **kwargs)
    ax.set_xscale(scale)
    ax.set_xlim(v_min, v_max)
    if flip_axis:
        ax.set_ylim(z_grid.max(), z_grid.min())
    else:
        ax.set_ylim(z_grid.min(), z_grid.max())
    ax.set_ylabel(y_label)

    return ax


def plotLayer(
    sig, LocSigZ, xscale="log", ax=None, showlayers=False, xlim=None, **kwargs
):
    """*plotLayer* has been deprecated, please use :func:`plot_1d_layer_model`"""
    warnings.warn(
        "plotLayer has been deprecated, please use plot_1d_layer_model",
        DeprecationWarning,
    )
    thicknesses = np.diff(LocSigZ)
    z0 = LocSigZ[0]
    ax = plot_1d_layer_model(
        thicknesses,
        sig,
        z0=z0,
        scale=xscale,
        ax=ax,
        show_layers=showlayers,
        plot_elevation=False,
    )
    if xlim is not None:
        ax.xlim(0.5 * sig.min(), 2 * sig.max())
    return ax



import numpy as np

from .simulation import BaseSimulation, BaseTimeSimulation
from .utils import mkvc, validate_type


class Fields:
    """Fancy Field Storage
    .. code::python
        fields = Fields(
            simulation=simulation, knownFields={"phi": "CC"}
        )
        fields[:,'phi'] = phi
        print(fields[src0,'phi'])
    """

    _dtype = float
    _knownFields = {}
    _aliasFields = {}

    def __init__(
        self, simulation, knownFields=None, aliasFields=None, dtype=None, **kwargs
    ):
        super().__init__(**kwargs)
        self.simulation = simulation

        if knownFields is not None:
            knownFields = validate_type("knownFields", knownFields, dict, cast=False)
            self._knownFields = knownFields
        if aliasFields is not None:
            aliasFields = validate_type("aliasFields", aliasFields, dict, cast=False)
            self._aliasFields = aliasFields
        if dtype is not None:
            self._dtype = dtype

        # check overlapping fields
        if any(key in self.aliasFields for key in self.knownFields):
            raise KeyError(
                "Aliased fields and Known Fields have overlapping definitions."
            )

        self._fields = {}
        self.startup()

    @property
    def simulation(self):
        """The simulation object that created these fields

        Returns
        -------
        SimPEG.simulation.BaseSimulation
        """
        return self._simulation

    @simulation.setter
    def simulation(self, value):
        self._simulation = validate_type(
            "simulation", value, BaseSimulation, cast=False
        )

    @property
    def knownFields(self):
        """The known fields of this object.

        The dictionary representing the known fields and their locations on the simulation
        mesh. The keys are the names of the fields, and the values are the location on
        the mesh.

        >>> fields.knownFields
        {'e': 'E', 'phi': 'CC'}

        Would represent that the `e` field and `phi` fields are known, and they are
        located on the mesh edges and cell centers, respectively.

        Returns
        -------
        dict
            They keys are the field names and the values are the field locations.
        """
        return self._knownFields

    @property
    def aliasFields(self):
        """The aliased fields of this object.

        The dictionary representing the aliased fields that can be accessed on this
        object. The keys are the names of the fields, and the values are a list of the
        known field, the aliased field's location on the mesh, and a function that goes
        from the known field to the aliased field.

        >>> fields.aliasFields
        {'b': ['e', 'F', '_e']}

        Would represent that the `e` field and `phi` fields are known, and they are
        located on the mesh edges and cell centers, respectively.

        Returns
        -------
        dict of {str: list}
            They keys are the field names and the values are list consiting of the
            field's alias, it's location on the mesh, and the function (or the name of
            it) to create it from the aliased field.
        """
        return self._aliasFields

    @property
    def dtype(self):
        """The data type of the storage matrix

        Returns
        -------
        dtype or dict of {str : dtype}
        """
        return self._dtype

    @property
    def knownFields(self):
        """Fields known to this object."""
        return self._knownFields

    @property
    def mesh(self):
        return self.simulation.mesh

    @property
    def survey(self):
        return self.simulation.survey

    def startup(self):
        pass

    @property
    def approxSize(self):
        """The approximate cost to storing all of the known fields."""
        sz = 0.0
        for f in self.knownFields:
            loc = self.knownFields[f]
            sz += np.array(self._storageShape(loc)).prod() * 8.0 / (1024**2)
        return "{0:e} MB".format(sz)

    def _storageShape(self, loc):
        n_fields = self.survey._n_fields

        nP = {
            "CC": self.mesh.nC,
            "N": self.mesh.nN,
            "F": self.mesh.nF,
            "E": self.mesh.nE,
        }[loc]

        return (nP, n_fields)

    def _initStore(self, name):
        if name in self._fields:
            return self._fields[name]

        assert name in self.knownFields, "field name is not known."

        loc = self.knownFields[name]

        if isinstance(self.dtype, dict):
            dtype = self.dtype[name]
        else:
            dtype = self.dtype

        # field = zarr.create(self._storageShape(loc), dtype=dtype)
        field = np.zeros(self._storageShape(loc), dtype=dtype)

        self._fields[name] = field

        return field

    def _srcIndex(self, srcTestList):
        if type(srcTestList) is slice:
            ind = srcTestList
        else:
            ind = self.survey.get_source_indices(srcTestList)
        return ind

    def _nameIndex(self, name, accessType):
        if type(name) is slice:
            assert name == slice(
                None, None, None
            ), "Fancy field name slicing is not supported... yet."
            name = None

        if name is None:
            return
        if accessType == "set" and name not in self.knownFields:
            if name in self.aliasFields:
                raise KeyError(
                    "Invalid field name ({0!s}) for setter, you can't "
                    "set an aliased property".format(name)
                )
            else:
                raise KeyError("Invalid field name ({0!s}) for setter".format(name))

        elif accessType == "get" and (
            name not in self.knownFields and name not in self.aliasFields
        ):
            raise KeyError("Invalid field name ({0!s}) for getter".format(name))
        return name

    def _index_name_srclist_from_key(self, key, accessType):
        if not isinstance(key, tuple):
            key = (key,)
        if len(key) == 1:
            key += (None,)

        assert len(key) == 2, "must be [Src, fieldName]"

        srcTestList, name = key
        name = self._nameIndex(name, accessType)
        ind = self._srcIndex(srcTestList)
        if isinstance(srcTestList, slice):
            srcTestList = self.survey.source_list[srcTestList]
        return ind, name, srcTestList

    def __setitem__(self, key, value):
        ind, name, src_list = self._index_name_srclist_from_key(key, "set")
        if name is None:
            assert isinstance(
                value, dict
            ), "New fields must be a dictionary, if field is not specified."
            newFields = value
        elif name in self.knownFields:
            newFields = {name: value}
        else:
            raise Exception("Unknown setter")

        for name in newFields:
            field = self._initStore(name)
            self._setField(field, newFields[name], name, ind)

    def __getitem__(self, key):
        ind, name, src_list = self._index_name_srclist_from_key(key, "get")
        if name is None:
            out = {}
            for name in self._fields:
                out[name] = self._getField(name, ind, src_list)
            return out
        return self._getField(name, ind, src_list)

    def _setField(self, field, val, name, ind):
        if isinstance(val, np.ndarray) and (
            field.shape[0] == field.size or val.ndim == 1
        ):
            val = mkvc(val, 2)
        field[:, ind] = val

    def _getField(self, name, ind, src_list):
        # ind will always be an list, thus the output will always
        # be (len(fields), n_inds)
        if name in self._fields:
            out = self._fields[name][:, ind]
        else:
            # Aliased fields
            alias, loc, func = self.aliasFields[name]

            if isinstance(func, str):
                assert hasattr(self, func), (
                    "The alias field function is a string, but it does not "
                    "exist in the Fields class."
                )
                func = getattr(self, func)
            if not isinstance(src_list, list):
                src_list = [src_list]
            out = func(self._fields[alias][:, ind], src_list)
        # if out.shape[0] == out.size or out.ndim == 1:
        #     out = mkvc(out, 2)
        return out

    def __contains__(self, other):
        if other in self.aliasFields:
            other = self.aliasFields[other][0]
        return self._fields.__contains__(other)


class TimeFields(Fields):
    """Fancy Field Storage for time domain problems
    .. code:: python

        fields = TimeFields(simulation=simulation, knownFields={'phi':'CC'})
        fields[:,'phi', timeInd] = phi
        print(fields[src0,'phi'])
    """

    @property
    def simulation(self):
        """The simulation object that created these fields

        Returns
        -------
        SimPEG.simulation.BaseTimeSimulation
        """
        return self._simulation

    @simulation.setter
    def simulation(self, value):
        self._simulation = validate_type(
            "simulation", value, BaseTimeSimulation, cast=False
        )

    def _storageShape(self, loc):
        nP = {
            "CC": self.mesh.nC,
            "N": self.mesh.nN,
            "F": self.mesh.nF,
            "E": self.mesh.nE,
        }[loc]
        nSrc = self.survey.nSrc
        nT = self.simulation.nT + 1
        return (nP, nSrc, nT)

    def _index_name_srclist_from_key(self, key, accessType):
        if not isinstance(key, tuple):
            key = (key,)
        if len(key) == 1:
            key += (None,)
        if len(key) == 2:
            key += (slice(None, None, None),)

        assert len(key) == 3, "must be [Src, fieldName, times]"

        srcTestList, name, timeInd = key

        name = self._nameIndex(name, accessType)
        srcInd = self._srcIndex(srcTestList)
        if isinstance(srcTestList, slice):
            srcTestList = self.survey.source_list[srcTestList]
        return (srcInd, timeInd), name, srcTestList

    def _correctShape(self, name, ind, deflate=False):
        srcInd, timeInd = ind
        if name in self.knownFields:
            loc = self.knownFields[name]
        else:
            loc = self.aliasFields[name][1]
        nP, total_nSrc, total_nT = self._storageShape(loc)
        nSrc = np.ones(total_nSrc, dtype=bool)[srcInd].sum()
        nT = np.ones(total_nT, dtype=bool)[timeInd].sum()
        shape = nP, nSrc, nT
        if deflate:
            shape = tuple([s for s in shape if s > 1])
        if len(shape) == 1:
            shape = shape + (1,)
        return shape

    def _setField(self, field, val, name, ind):
        srcInd, timeInd = ind
        shape = self._correctShape(name, ind)
        if isinstance(val, np.ndarray) and val.size == 1:
            val = val[0]
        if np.isscalar(val):
            field[:, srcInd, timeInd] = val
            return
        if val.size != np.array(shape).prod():
            raise ValueError("Incorrect size for data.")
        correctShape = field[:, srcInd, timeInd].shape
        field[:, srcInd, timeInd] = val.reshape(correctShape, order="F")

    def _getField(self, name, ind, src_list):
        srcInd, timeInd = ind

        if name in self._fields:
            out = self._fields[name][:, srcInd, timeInd]
        else:
            # Aliased fields
            alias, loc, func = self.aliasFields[name]
            if isinstance(func, str):
                assert hasattr(self, func), (
                    "The alias field function is a string, but it does "
                    "not exist in the Fields class."
                )
                func = getattr(self, func)
            pointerFields = self._fields[alias][:, srcInd, timeInd]
            pointerShape = self._correctShape(alias, ind)
            pointerFields = pointerFields.reshape(pointerShape, order="F")

            # First try to return the function as three arguments (without timeInd)
            if timeInd == slice(None, None, None):
                try:
                    # assume it will take care of integrating over all times
                    return func(pointerFields, srcInd)
                except TypeError:
                    pass

            timeII = np.arange(self.simulation.nT + 1)[timeInd]
            if not isinstance(src_list, list):
                src_list = [src_list]

            if timeII.size == 1:
                pointerShapeDeflated = self._correctShape(alias, ind, deflate=True)
                pointerFields = pointerFields.reshape(pointerShapeDeflated, order="F")
                out = func(pointerFields, src_list, timeII)
            else:  # loop over the time steps
                nT = pointerShape[2]
                out = list(range(nT))
                for i, TIND_i in enumerate(timeII):
                    fieldI = pointerFields[:, :, i]
                    if fieldI.shape[0] == fieldI.size:
                        fieldI = mkvc(fieldI, 2)
                    out[i] = func(fieldI, src_list, TIND_i)
                    if out[i].ndim == 1:
                        out[i] = out[i][:, np.newaxis, np.newaxis]
                    elif out[i].ndim == 2:
                        out[i] = out[i][:, :, np.newaxis]
                out = np.concatenate(out, axis=2)

        shape = self._correctShape(name, ind, deflate=True)
        return out.reshape(shape, order="F")



import numpy as np

from ..data_misfit import L2DataMisfit
from ..fields import Fields
from ..utils import mkvc, sdiag
from .utils import compute
import dask.array as da
from scipy.sparse import csr_matrix as csr
from dask import delayed


def dask_call(self, m, f=None):
    """
    Distributed :obj:`simpeg.data_misfit.L2DataMisfit.__call__`
    """
    R = self.W * self.residual(m, f=f)
    phi_d = 0.5 * da.dot(R, R)
    if not isinstance(phi_d, np.ndarray):
        return compute(self, phi_d)
    return phi_d

L2DataMisfit.__call__ = dask_call


def dask_deriv(self, m, f=None):
    """
    Distributed :obj:`simpeg.data_misfit.L2DataMisfit.deriv`
    """
    mapping_deriv = self.model_map.deriv(m)
    if getattr(self, "model_map", None) is not None:
        m = self.model_map @ m

    wtw_d = self.W.diagonal() ** 2.0 * self.residual(m, f=f)
    Jtvec = compute(self, self.simulation.Jtvec(m, wtw_d))

    if getattr(self, "model_map", None) is not None:
        Jtjvec_dmudm = delayed(csr.dot)(Jtvec, mapping_deriv)
        h_vec = da.from_delayed(
            Jtjvec_dmudm, dtype=float, shape=[mapping_deriv.shape[1]]
        )
        if not isinstance(h_vec, np.ndarray):
            return compute(self, h_vec)
        return h_vec

    if not isinstance(Jtvec, np.ndarray):
        return compute(self, Jtvec)
    return Jtvec


L2DataMisfit.deriv = dask_deriv


def dask_deriv2(self, m, v, f=None):
    """
    Distributed :obj:`simpeg.data_misfit.L2DataMisfit.deriv2`
    """
    mapping_deriv = self.model_map.deriv(m)
    if getattr(self, "model_map", None) is not None:
        m = self.model_map @ m
        v = mapping_deriv @ v

    jvec = compute(self, self.simulation.Jvec(m, v))
    w_jvec = self.W.diagonal() ** 2.0 * jvec
    jtwjvec = compute(self, self.simulation.Jtvec(m, w_jvec))

    if getattr(self, "model_map", None) is not None:
        Jtjvec_dmudm = delayed(csr.dot)(jtwjvec, mapping_deriv)
        h_vec = da.from_delayed(
            Jtjvec_dmudm, dtype=float, shape=[mapping_deriv.shape[1]]
        )
        if not isinstance(h_vec, np.ndarray):
            return compute(self, h_vec)
        return h_vec

    if not isinstance(jtwjvec, np.ndarray):
        return compute(self, jtwjvec)
    return jtwjvec


L2DataMisfit.deriv2 = dask_deriv2


def dask_residual(self, m, f=None):
    if self.data is None:
        raise Exception("data must be set before a residual can be calculated.")

    if isinstance(f, Fields) or f is None:
        return self.simulation.residual(m, self.data.dobs, f=f)
    elif f.shape == self.data.dobs.shape:
        return mkvc(f - self.data.dobs)
    else:
        raise Exception(f"Attribute f must be or type {Fields}, numpy.array or None.")


L2DataMisfit.residual = dask_residual


import numpy as np

import discretize
from SimPEG import maps, mkvc, utils, Data
from ....utils import unpack_widths
from ..receivers import (
    PointNaturalSource,
    Point3DTipper,
)
from ..survey import Survey
from ..sources import PlanewaveXYPrimary, Planewave
from ..simulation import Simulation3DPrimarySecondary
from .data_utils import appResPhs

np.random.seed(1100)
# Define the tolerances
TOLr = 5e-2
TOLp = 5e-2


def getAppResPhs(NSEMdata, survey):
    NSEMdata = Data(dobs=NSEMdata, survey=survey)
    # Make impedance
    zList = []
    for src in survey.source_list:
        zc = [src.frequency]
        for rx in src.receiver_list:
            if "imag" in rx.component:
                m = 1j
            else:
                m = 1
            zc.append(m * NSEMdata[src, rx])
        zList.append(zc)
    return [
        appResPhs(zList[i][0], np.sum(zList[i][1:3])) for i in np.arange(len(zList))
    ]


def setup1DSurvey(sigmaHalf, tD=False, structure=False):
    # Frequency
    num_frequencies = 33
    freqs = np.logspace(3, -3, num_frequencies)
    # Make the mesh
    ct = 5
    air = unpack_widths([(ct, 25, 1.3)])
    # coreT0 = unpack_widths([(ct,15,1.2)])
    # coreT1 = np.kron(unpack_widths([(coreT0[-1],15,1.3)]),np.ones((7,)))
    core = np.concatenate(
        (
            np.kron(unpack_widths([(ct, 15, -1.2)]), np.ones((10,))),
            unpack_widths([(ct, 20)]),
        )
    )
    bot = unpack_widths([(core[0], 20, -1.3)])
    x0 = -np.array([np.sum(np.concatenate((core, bot)))])
    m1d = discretize.TensorMesh([np.concatenate((bot, core, air))], x0=x0)
    # Make the model
    sigma = np.zeros(m1d.nC) + sigmaHalf
    sigma[m1d.gridCC > 0] = 1e-8
    sigmaBack = sigma.copy()
    # Add structure
    if structure:
        shallow = (m1d.gridCC < -200) * (m1d.gridCC > -600)
        deep = (m1d.gridCC < -3000) * (m1d.gridCC > -5000)
        sigma[shallow] = 1
        sigma[deep] = 0.1

    receiver_list = []
    for _ in range(len(["z1d", "z1d"])):
        receiver_list.append(
            PointNaturalSource(mkvc(np.array([0.0]), 2).T, component="real")
        )
        receiver_list.append(
            PointNaturalSource(mkvc(np.array([0.0]), 2).T, component="imag")
        )
    # Source list
    source_list = []
    for freq in freqs:
        source_list.append(PlanewaveXYPrimary(receiver_list, freq))

    survey = Survey(source_list)
    return (survey, sigma, sigmaBack, m1d)


def setup1DSurveyElectricMagnetic(sigmaHalf, tD=False, structure=False):
    # Frequency
    nFreq = 33
    frequencies = np.logspace(3, -3, nFreq)
    # Make the mesh
    ct = 5
    air = unpack_widths([(ct, 25, 1.3)])
    # coreT0 = unpack_widths([(ct,15,1.2)])
    # coreT1 = np.kron(unpack_widths([(coreT0[-1],15,1.3)]),np.ones((7,)))
    core = np.concatenate(
        (
            np.kron(unpack_widths([(ct, 15, -1.2)]), np.ones((10,))),
            unpack_widths([(ct, 20)]),
        )
    )
    bot = unpack_widths([(core[0], 20, -1.3)])
    x0 = -np.array([np.sum(np.concatenate((core, bot)))])
    m1d = discretize.TensorMesh([np.concatenate((bot, core, air))], x0=x0)
    # Make the model
    sigma = np.zeros(m1d.nC) + sigmaHalf
    sigma[m1d.gridCC > 0] = 1e-8
    sigmaBack = sigma.copy()
    # Add structure
    if structure:
        shallow = (m1d.gridCC < -200) * (m1d.gridCC > -600)
        deep = (m1d.gridCC < -3000) * (m1d.gridCC > -5000)
        sigma[shallow] = 1
        sigma[deep] = 0.1

    rxList = []
    for _ in range(len(["z1d", "z1d"])):
        rxList.append(PointNaturalSource(mkvc(np.array([0.0]), 2).T, component="real"))
        rxList.append(PointNaturalSource(mkvc(np.array([0.0]), 2).T, component="imag"))
    # Source list
    # srcList = []
    src_list = [Planewave([], frequency=f) for f in frequencies]
    # if tD:
    #     for freq in freqs:
    #         srcList.append(Planewave_xy_1DhomotD(rxList, freq))
    # else:
    #     for freq in freqs:
    #         srcList.append(PlanewaveXYPrimary(rxList, freq))

    survey = Survey(src_list)
    return (survey, sigma, sigmaBack, m1d, frequencies)


def setupSimpegNSEM_tests_location_assign_list(
    inputSetup, freqs, comp="Imp", singleFreq=False, singleList=False
):
    rx_x, rx_y = np.meshgrid(np.linspace(-5000, 5000, 5), np.linspace(-5000, 5000, 5))

    rx_loc = np.hstack(
        (mkvc(rx_x, 2), mkvc(rx_y, 2), np.zeros((np.prod(rx_x.shape), 1)))
    )

    csx = 2000.0
    csz = 2000.0

    mesh = discretize.TensorMesh(
        [
            [(csx, 1, -3), (csx, 6), (csx, 1, 3)],
            [(csx, 1, -3), (csx, 6), (csx, 1, 3)],
            [(csz, 1, -3), (csz, 6), (csz, 1, 3)],
        ],
        x0="CCC",
    )

    active = mesh.gridCC[:, 2] < 50
    sig = np.ones(mesh.nC) * 0.01
    sig[~active] = 1e-8
    model_true = np.log(sig)

    print(model_true.shape, mesh.nE * 2, active.sum())

    rx_loc[:, 2] = -50

    rxList = []
    if comp == "All":
        rx_type_list = ["xx", "xy", "yx", "yy", "zx", "zy"]
    elif comp == "Imp":
        rx_type_list = ["xx", "xy", "yx", "yy"]
    elif comp == "Tip":
        rx_type_list = ["zx", "zy"]
    elif comp == "Res":
        rx_type_list = ["yx", "xy"]
    else:
        rx_type_list = [comp]

    for rx_type in rx_type_list:
        if rx_type in ["xx", "xy", "yx", "yy"]:
            if comp == "Res":
                if singleList:
                    rxList.append(
                        PointNaturalSource(
                            locations=[rx_loc],
                            orientation=rx_type,
                            component="apparent_resistivity",
                        )
                    )
                    rxList.append(
                        PointNaturalSource(
                            locations=[rx_loc], orientation=rx_type, component="phase"
                        )
                    )
                else:
                    rxList.append(
                        PointNaturalSource(
                            locations=[rx_loc, rx_loc],
                            orientation=rx_type,
                            component="apparent_resistivity",
                        )
                    )
                    rxList.append(
                        PointNaturalSource(
                            locations=[rx_loc, rx_loc],
                            orientation=rx_type,
                            component="phase",
                        )
                    )
            else:
                rxList.append(
                    PointNaturalSource(
                        orientation=rx_type, component="real", locations=[rx_loc]
                    )
                )
                rxList.append(
                    PointNaturalSource(
                        orientation=rx_type, component="imag", locations=[rx_loc]
                    )
                )
        if rx_type in ["zx", "zy"]:
            rxList.append(
                Point3DTipper(orientation=rx_type, component="real", locations=[rx_loc])
            )
            rxList.append(
                Point3DTipper(orientation=rx_type, component="imag", locations=[rx_loc])
            )

    srcList = []
    if singleFreq:
        # srcList.append(PlanewaveXYPrimary(rxList, singleFreq, sigma_primary=sigBG))
        srcList.append(PlanewaveXYPrimary(rxList, singleFreq))
    else:
        for freq in freqs:
            # srcList.append(PlanewaveXYPrimary(rxList, freq, sigma_primary=sigBG))
            srcList.append(PlanewaveXYPrimary(rxList, freq))

    # Survey MT
    survey_ns = Survey(srcList)
    survey_ns.m_true = model_true

    # write out the true model
    # discretize.TensorMesh.write_UBC(mesh,'Mesh-pre.msh', models={'Sigma-pre.dat': np.exp(model_true)})
    # create background conductivity model
    sigma_back = 1e-2
    sigBG = np.zeros(mesh.nC) * sigma_back
    sigBG[~active] = 1e-8

    # Set the mapping
    actMap = maps.InjectActiveCells(
        mesh=mesh, indActive=active, valInactive=np.log(1e-8)
    )
    mapping = maps.ExpMap(mesh) * actMap
    # print(survey_ns.source_list)
    # # Setup the problem object
    sim = Simulation3DPrimarySecondary(
        mesh, survey=survey_ns, sigmaPrimary=sigBG, sigmaMap=mapping
    )

    # create the test model
    block = [csx * np.r_[-3, 3], csx * np.r_[-3, 3], csz * np.r_[-6, -1]]

    block_sigma = 3e-1

    block_inds = (
        (mesh.gridCC[:, 0] >= block[0].min())
        & (mesh.gridCC[:, 0] <= block[0].max())
        & (mesh.gridCC[:, 1] >= block[1].min())
        & (mesh.gridCC[:, 1] <= block[1].max())
        & (mesh.gridCC[:, 2] >= block[2].min())
        & (mesh.gridCC[:, 2] <= block[2].max())
    )

    m = model_true.copy()
    m[block_inds] = np.log(block_sigma)
    m = m[active]

    # f = sim.fields(m)

    # TOLr = 5e-2
    # TOL = 1e-4
    # FLR = 1e-20

    # w = np.random.rand(len(m),)
    # v = np.random.rand(sim.survey.nD,)
    # vJw = v.ravel().dot(sim.Jvec(m, w, f))
    # wJtv = w.ravel().dot(sim.Jtvec(m, v, f))

    return (m, sim)


def setupSimpegNSEM_PrimarySecondary(inputSetup, freqs, comp="Imp", singleFreq=False):
    rx_x, rx_y = np.meshgrid(np.linspace(-5000, 5000, 5), np.linspace(-5000, 5000, 5))

    rx_loc = np.hstack(
        (mkvc(rx_x, 2), mkvc(rx_y, 2), np.zeros((np.prod(rx_x.shape), 1)))
    )

    csx = 2000.0
    csz = 2000.0

    mesh = discretize.TensorMesh(
        [
            [(csx, 1, -3), (csx, 6), (csx, 1, 3)],
            [(csx, 1, -3), (csx, 6), (csx, 1, 3)],
            [(csz, 1, -3), (csz, 6), (csz, 1, 3)],
        ],
        x0="CCC",
    )

    active = mesh.gridCC[:, 2] < 50
    sig = np.ones(mesh.nC) * 0.01
    sig[~active] = 1e-8
    model_true = np.log(sig)

    print(model_true.shape, mesh.nE * 2, active.sum())

    rx_loc[:, 2] = -50

    rxList = []
    if comp == "All":
        rx_type_list = ["xx", "xy", "yx", "yy", "zx", "zy"]
    elif comp == "Imp":
        rx_type_list = ["xx", "xy", "yx", "yy"]
    elif comp == "Tip":
        rx_type_list = ["zx", "zy"]
    elif comp == "Res":
        rx_type_list = ["yx", "xy"]
    else:
        rx_type_list = [comp]

    for rx_type in rx_type_list:
        if rx_type in ["xx", "xy", "yx", "yy"]:
            if comp == "Res":
                rxList.append(
                    PointNaturalSource(
                        locations=rx_loc,
                        orientation=rx_type,
                        component="apparent_resistivity",
                    )
                )
                rxList.append(
                    PointNaturalSource(
                        locations=rx_loc, orientation=rx_type, component="phase"
                    )
                )
            else:
                rxList.append(PointNaturalSource(rx_loc, rx_type, "real"))
                rxList.append(PointNaturalSource(rx_loc, rx_type, "imag"))
        if rx_type in ["zx", "zy"]:
            rxList.append(Point3DTipper(rx_loc, rx_type, "real"))
            rxList.append(Point3DTipper(rx_loc, rx_type, "imag"))

    srcList = []
    if singleFreq:
        # srcList.append(PlanewaveXYPrimary(rxList, singleFreq, sigma_primary=sigBG))
        srcList.append(PlanewaveXYPrimary(rxList, singleFreq))
    else:
        for freq in freqs:
            # srcList.append(PlanewaveXYPrimary(rxList, freq, sigma_primary=sigBG))
            srcList.append(PlanewaveXYPrimary(rxList, freq))

    # Survey MT
    survey_ns = Survey(srcList)
    survey_ns.m_true = model_true

    # write out the true model
    # discretize.TensorMesh.write_UBC(mesh,'Mesh-pre.msh', models={'Sigma-pre.dat': np.exp(model_true)})
    # create background conductivity model
    sigma_back = 1e-2
    sigBG = np.zeros(mesh.nC) * sigma_back
    sigBG[~active] = 1e-8

    # Set the mapping
    actMap = maps.InjectActiveCells(
        mesh=mesh, indActive=active, valInactive=np.log(1e-8)
    )
    mapping = maps.ExpMap(mesh) * actMap
    # print(survey_ns.source_list)
    # # Setup the problem object
    sim = Simulation3DPrimarySecondary(
        mesh, survey=survey_ns, sigmaPrimary=sigBG, sigmaMap=mapping
    )

    # create the test model
    block = [csx * np.r_[-3, 3], csx * np.r_[-3, 3], csz * np.r_[-6, -1]]

    block_sigma = 3e-1

    block_inds = (
        (mesh.gridCC[:, 0] >= block[0].min())
        & (mesh.gridCC[:, 0] <= block[0].max())
        & (mesh.gridCC[:, 1] >= block[1].min())
        & (mesh.gridCC[:, 1] <= block[1].max())
        & (mesh.gridCC[:, 2] >= block[2].min())
        & (mesh.gridCC[:, 2] <= block[2].max())
    )

    m = model_true.copy()
    m[block_inds] = np.log(block_sigma)
    m = m[active]

    # f = sim.fields(m)

    # TOLr = 5e-2
    # TOL = 1e-4
    # FLR = 1e-20

    # w = np.random.rand(len(m),)
    # v = np.random.rand(sim.survey.nD,)
    # vJw = v.ravel().dot(sim.Jvec(m, w, f))
    # wJtv = w.ravel().dot(sim.Jtvec(m, v, f))

    return (m, sim)


def setupSimpegNSEM_ePrimSec(inputSetup, comp="Imp", singleFreq=False, expMap=True):
    M, freqs, sig, sigBG, rx_loc = inputSetup
    # Make a receiver list
    receiver_list = []
    if comp == "All":
        rx_type_list = ["xx", "xy", "yx", "yy", "zx", "zy"]
    elif comp == "Imp":
        rx_type_list = ["xx", "xy", "yx", "yy"]
    elif comp == "Tip":
        rx_type_list = ["zx", "zy"]
    elif comp == "Res":
        rx_type_list = ["yx", "xy"]
    else:
        rx_type_list = [comp]

    for rx_type in rx_type_list:
        if rx_type in ["xx", "xy", "yx", "yy"]:
            if comp == "Res":
                receiver_list.append(
                    PointNaturalSource(
                        locations_e=rx_loc,
                        locations_h=rx_loc,
                        orientation=rx_type,
                        component="apparent_resistivity",
                    )
                )
                receiver_list.append(
                    PointNaturalSource(
                        locations_e=rx_loc,
                        locations_h=rx_loc,
                        orientation=rx_type,
                        component="phase",
                    )
                )
            else:
                receiver_list.append(PointNaturalSource(rx_loc, rx_type, "real"))
                receiver_list.append(PointNaturalSource(rx_loc, rx_type, "imag"))
        if rx_type in ["zx", "zy"]:
            receiver_list.append(Point3DTipper(rx_loc, rx_type, "real"))
            receiver_list.append(Point3DTipper(rx_loc, rx_type, "imag"))

    # Source list
    source_list = []

    if singleFreq:
        source_list.append(PlanewaveXYPrimary(receiver_list, singleFreq))
    else:
        for freq in freqs:
            source_list.append(PlanewaveXYPrimary(receiver_list, freq))
    # Survey NSEM
    survey = Survey(source_list)

    # Setup the problem object
    if expMap:
        problem = Simulation3DPrimarySecondary(
            M, survey=survey, sigmaPrimary=np.log(sigBG), sigmaMap=maps.ExpMap(M)
        )
        problem.model = np.log(sig)
    else:
        problem = Simulation3DPrimarySecondary(
            M, survey=survey, sigmaPrimary=sigBG, sigmaMap=maps.IdentityMap(M)
        )
        problem.model = sig
    problem.verbose = False
    try:
        from pymatsolver import Pardiso

        problem.solver = Pardiso
    except ImportError:
        pass

    return (survey, problem)


def getInputs():
    """
    Function that returns Mesh, freqs, rx_loc, elev.
    """
    # Make a mesh
    M = discretize.TensorMesh(
        [
            [(200, 6, -1.5), (200.0, 4), (200, 6, 1.5)],
            [(200, 6, -1.5), (200.0, 4), (200, 6, 1.5)],
            [(200, 8, -1.5), (200.0, 8), (200, 8, 1.5)],
        ],
        x0=["C", "C", "C"],
    )  # Setup the model
    # Set the frequencies
    freqs = np.logspace(1, -3, 5)
    elev = 0

    # Setup the the survey object
    # Receiver locations
    rx_x, rx_y = np.meshgrid(np.arange(-350, 350, 200), np.arange(-350, 350, 200))
    rx_loc = np.hstack(
        (mkvc(rx_x, 2), mkvc(rx_y, 2), elev + np.zeros((np.prod(rx_x.shape), 1)))
    )

    return M, freqs, rx_loc, elev


def random(conds):
    """Returns a random model based on the inputs"""
    M, freqs, rx_loc, elev = getInputs()

    # Backround
    sigBG = np.ones(M.nC) * conds
    # Add randomness to the model (10% of the value).
    sig = np.exp(np.log(sigBG) + np.random.randn(M.nC) * (conds) * 1e-1)

    return (M, freqs, sig, sigBG, rx_loc)


def halfSpace(conds):
    """Returns a halfspace model based on the inputs"""
    M, freqs, rx_loc, elev = getInputs()

    # Model
    ccM = M.gridCC
    # conds = [1e-2]
    groundInd = ccM[:, 2] < elev
    sig = np.zeros(M.nC) + 1e-8
    sig[groundInd] = conds
    # Set the background, not the same as the model
    sigBG = np.zeros(M.nC) + 1e-8
    sigBG[groundInd] = conds

    return (M, freqs, sig, sigBG, rx_loc)


def blockInhalfSpace(conds):
    """Returns a block in a halfspace model based on the inputs"""
    M, freqs, rx_loc, elev = getInputs()

    # Model
    ccM = M.gridCC
    # conds = [1e-2]
    groundInd = ccM[:, 2] < elev
    sig = utils.model_builder.defineBlock(
        M.gridCC, np.array([-1000, -1000, -1500]), np.array([1000, 1000, -1000]), conds
    )
    sig[~groundInd] = 1e-8
    # Set the background, not the same as the model
    sigBG = np.zeros(M.nC) + 1e-8
    sigBG[groundInd] = conds[1]

    return (M, freqs, sig, sigBG, rx_loc)


def twoLayer(conds):
    """Returns a 2 layer model based on the conductivity values given"""
    M, freqs, rx_loc, elev = getInputs()

    # Model
    ccM = M.gridCC
    groundInd = ccM[:, 2] < elev
    botInd = ccM[:, 2] < -3000
    sig = np.zeros(M.nC) + 1e-8
    sig[groundInd] = conds[1]
    sig[botInd] = conds[0]
    # Set the background, not the same as the model
    sigBG = np.zeros(M.nC) + 1e-8
    sigBG[groundInd] = conds[1]

    return (M, freqs, sig, sigBG, rx_loc)



import numpy as np
import scipy.constants as constants
from geoana.kernels import prism_fz, prism_fzx, prism_fzy, prism_fzz
from scipy.constants import G as NewtG

from SimPEG import props
from SimPEG.utils import mkvc, sdiag

from ...base import BasePDESimulation
from ..base import BaseEquivalentSourceLayerSimulation, BasePFSimulation


class Simulation3DIntegral(BasePFSimulation):
    """
    Gravity simulation in integral form.

    """

    rho, rhoMap, rhoDeriv = props.Invertible("Density")

    def __init__(self, mesh, rho=None, rhoMap=None, **kwargs):
        super().__init__(mesh, model_map=rhoMap, **kwargs)
        self.rho = rho
        self.rhoMap = rhoMap

    def fields(self, m):
        self.model = m

        if self.store_sensitivities is None:
            # Compute the linear operation without forming the full dense Jmatrix
            fields = mkvc(self.linear_operator())
        else:
            fields = self.Jmatrix @ (self.rho).astype(np.float32)

        return np.asarray(fields)

    def getJtJdiag(self, m, W=None, f=None):
        """
        Return the diagonal of JtJ
        """
        self.model = m

        if W is None:
            W = np.ones(self.survey.nD)
        else:
            W = W.diagonal() ** 2
        if getattr(self, "_gtg_diagonal", None) is None:
            diag = np.zeros(self.Jmatrix.shape[1])
            for i in range(len(W)):
                diag += W[i] * (self.Jmatrix[i] * self.Jmatrix[i])
            self._gtg_diagonal = diag
        else:
            diag = self._gtg_diagonal
        return mkvc((sdiag(np.sqrt(diag)) @ self.rhoDeriv).power(2).sum(axis=0))

    def getJ(self, m, f=None):
        """
        Sensitivity matrix
        """
        return self.Jmatrix.dot(self.rhoDeriv)

    def Jvec(self, m, v, f=None):
        """
        Sensitivity times a vector
        """
        dmu_dm_v = self.rhoDeriv @ v
        return self.Jmatrix @ dmu_dm_v.astype(np.float32)

    def Jtvec(self, m, v, f=None):
        """
        Sensitivity transposed times a vector
        """
        Jtvec = self.Jmatrix.T @ v.astype(np.float32)
        return np.asarray(self.rhoDeriv.T @ Jtvec)

    @property
    def gtg_diagonal(self):
        """
        Diagonal of GtG
        """
        if getattr(self, "_gtg_diagonal", None) is None:
            return None

        return self._gtg_diagonal

    def evaluate_integral(self, receiver_location, components):
        """
        Compute the forward linear relationship between the model and the physics at a point
        and for all components of the survey.

        :param numpy.ndarray receiver_location:  array with shape (n_receivers, 3)
            Array of receiver locations as x, y, z columns.
        :param list[str] components: List of gravity components chosen from:
            'gx', 'gy', 'gz', 'gxx', 'gxy', 'gxz', 'gyy', 'gyz', 'gzz', 'guv'
        :param float tolerance: Small constant to avoid singularity near nodes and edges.
        :rtype numpy.ndarray: rows
        :returns: ndarray with shape (n_components, n_cells)
            Dense array mapping of the contribution of all active cells to data components::

                rows =
                    g_1 = [g_1x g_1y g_1z]
                    g_2 = [g_2x g_2y g_2z]
                           ...
                    g_c = [g_cx g_cy g_cz]

        """
        dr = self._nodes - receiver_location
        dx = dr[..., 0]
        dy = dr[..., 1]
        dz = dr[..., 2]

        node_evals = {}
        if "gx" in components:
            node_evals["gx"] = prism_fz(dy, dz, dx)
        if "gy" in components:
            node_evals["gy"] = prism_fz(dz, dx, dy)
        if "gz" in components:
            node_evals["gz"] = prism_fz(dx, dy, dz)
        if "gxy" in components:
            node_evals["gxy"] = prism_fzx(dy, dz, dx)
        if "gxz" in components:
            node_evals["gxz"] = prism_fzx(dx, dy, dz)
        if "gyz" in components:
            node_evals["gyz"] = prism_fzy(dx, dy, dz)
        if "gxx" in components or "guv" in components:
            node_evals["gxx"] = prism_fzz(dy, dz, dx)
        if "gyy" in components or "guv" in components:
            node_evals["gyy"] = prism_fzz(dz, dx, dy)
            if "guv" in components:
                node_evals["guv"] = (node_evals["gyy"] - node_evals["gxx"]) * 0.5
                # (NN - EE) / 2
        inside_adjust = False
        if "gzz" in components:
            if "gxx" not in node_evals or "gyy" not in node_evals:
                node_evals["gzz"] = prism_fzz(dx, dy, dz)
            else:
                inside_adjust = True
                # The below need to be adjusted for observation points within a cell.
                # because `gxx + gyy + gzz = -4 * pi * G * rho`
                # gzz = - gxx - gyy - 4 * np.pi * G * rho[in_cell]
                node_evals["gzz"] = -node_evals["gxx"] - node_evals["gyy"]

        rows = {}
        for component in set(components):
            vals = node_evals[component]
            if self._unique_inv is not None:
                vals = vals[self._unique_inv]
            cell_vals = (
                vals[0]
                - vals[1]
                - vals[2]
                + vals[3]
                - vals[4]
                + vals[5]
                + vals[6]
                - vals[7]
            )
            if inside_adjust and component == "gzz":
                # should subtract 4 * pi to the cell containing the observation point
                # just need a little logic to find the containing cell
                # cell_vals[inside_cell] += 4 * np.pi
                pass
            if self.store_sensitivities is None:
                rows[component] = cell_vals @ self.rho
            else:
                rows[component] = cell_vals
            if len(component) == 3:
                rows[component] *= constants.G * 1e12  # conversion for Eotvos
            else:
                rows[component] *= constants.G * 1e8  # conversion for mGal

        return np.stack([rows[component] for component in components])


class SimulationEquivalentSourceLayer(
    BaseEquivalentSourceLayerSimulation, Simulation3DIntegral
):
    """
    Equivalent source layer simulations

    Parameters
    ----------
    mesh : discretize.BaseMesh
        A 2D tensor or tree mesh defining discretization along the x and y directions
    cell_z_top : numpy.ndarray or float
        Define the elevations for the top face of all cells in the layer. If an array it should be the same size as
        the active cell set.
    cell_z_bottom : numpy.ndarray or float
        Define the elevations for the bottom face of all cells in the layer. If an array it should be the same size as
        the active cell set.
    """


class Simulation3DDifferential(BasePDESimulation):
    r"""Finite volume simulation class for gravity.

    Notes
    -----
    From Blakely (1996), the scalar potential :math:`\phi` outside the source region
    is obtained by solving a Poisson's equation:

    .. math::
        \nabla^2 \phi = 4 \pi \gamma \rho

    where :math:`\gamma` is the gravitational constant and :math:`\rho` defines the
    distribution of density within the source region.

    Applying the finite volumn method, we can solve the Poisson's equation on a
    3D voxel grid according to:

    .. math::
        \big [ \mathbf{D M_f D^T} \big ] \mathbf{u} = - \mathbf{M_c \, \rho}
    """

    rho, rhoMap, rhoDeriv = props.Invertible("Specific density (g/cc)")

    def __init__(self, mesh, rho=1.0, rhoMap=None, **kwargs):
        super().__init__(mesh, **kwargs)
        self.rho = rho
        self.rhoMap = rhoMap

        self._Div = self.mesh.face_divergence

    def getRHS(self):
        """Return right-hand side for the linear system"""
        Mc = self.Mcc
        rho = self.rho
        return -Mc * rho

    def getA(self):
        r"""
        GetA creates and returns the A matrix for the Gravity nodal problem

        The A matrix has the form:

        .. math ::

            \mathbf{A} =  \Div(\Mf Mui)^{-1}\Div^{T}
        """
        # Constructs A with 0 dirichlet
        if getattr(self, "_A", None) is None:
            self._A = self._Div * self.Mf * self._Div.T.tocsr()
        return self._A

    def fields(self, m=None):
        r"""Compute fields

        **INCOMPLETE**

        Parameters
        ----------
        m: (nP) np.ndarray
            The model

        Returns
        -------
        dict
            The fields
        """
        if m is not None:
            self.model = m

        A = self.getA()
        RHS = self.getRHS()

        Ainv = self.solver(A)
        u = Ainv * RHS

        gField = 4.0 * np.pi * NewtG * 1e8 * self._Div * u

        return {"G": gField, "u": u}



import numpy as np
from ...utils import mkvc
import discretize
import warnings


def edge_basis_function(t, a1, l1, h1, a2, l2, h2):
    """Edge basis function

    Parameters
    ----------
    t : float
        Parameterized distance along wire
    a1 : float
        Start of wire in first coordinate
    l1 : float
        Length of wire in first coordinate
    h1 : float
        Cell dimension in first coordinate
    a2 : float
        Start of wire in second coordinate
    l2 : float
        Length of wire in second coordinate
    h2 : float
        Cell dimension in second coordinate

    Returns
    -------
    (4) numpy.ndarray
        Edge weights
    """
    x1 = a1 + t * l1
    x2 = a2 + t * l2
    w0 = (1.0 - x1 / h1) * (1.0 - x2 / h2)
    w1 = (x1 / h1) * (1.0 - x2 / h2)
    w2 = (1.0 - x1 / h1) * (x2 / h2)
    w3 = (x1 / h1) * (x2 / h2)
    return np.r_[w0, w1, w2, w3]


def _simpsons_rule(a1, l1, h1, a2, l2, h2):
    """Return weights for Simpson's rule."""
    wl = edge_basis_function(0.0, a1, l1, h1, a2, l2, h2)
    wc = edge_basis_function(0.5, a1, l1, h1, a2, l2, h2)
    wr = edge_basis_function(1.0, a1, l1, h1, a2, l2, h2)
    return (wl + 4.0 * wc + wr) / 6.0


# TODO: Extend this when current is defined on cell-face
def getStraightLineCurrentIntegral(hx, hy, hz, ax, ay, az, bx, by, bz):
    """Compute straight current line integral

    Compute integral int(W . J dx^3) in brick of size hx x hy x hz
    where W denotes the 12 local bilinear edge basis functions
    and where J prescribes a unit line current
    between points (ax,ay,az) and (bx,by,bz).

    Parameters
    ----------
    hx : float
        x cell dimension
    hy : float
        y cell dimension
    hz : float
        z cell dimension
    ax : float
        x-location of start of wire segement
    ay :float
        y-location of start of wire segement
    az :float
        z-location of start of wire segement
    bx :float
        x-location of end of wire segement
    by :float
        y-location of end of wire segement
    bz :float
        z-location of end of wire segement

    Returns
    -------
    sx : float
        x weight
    sy : float
        y weight
    sz : flat
        z weight
    """

    # length of line segment
    lx = bx - ax
    ly = by - ay
    lz = bz - az

    # integration using Simpson's rule
    sx = _simpsons_rule(ay, ly, hy, az, lz, hz) * lx
    sy = _simpsons_rule(ax, lx, hx, az, lz, hz) * ly
    sz = _simpsons_rule(ax, lx, hx, ay, ly, hy) * lz

    return sx, sy, sz


def _findlast(x):
    """Returns last element

    Parameters
    ----------
    array_like of int

    Returns
    -------

    """
    if x.sum() == 0:
        return -1
    else:
        return np.arange(x.size)[x][-1]


def segmented_line_current_source_term(mesh, locs):
    """Calculate a source term for a line current source on a mesh

    Given a discretize mesh, compute the source vector for a unit current flowing
    along the segmented line path with vertices defined by `locs`.

    Parameters
    ----------
    mesh : discretize.TreeMesh or discretize.TensorMesh
        The Mesh (3D) for the system.
    locs : (n_points, 3) numpy.ndarray
        The array of locations of consecutive points along the polygonal path.

    Returns
    -------
    (mesh.n_edges) numpy.ndarray
        Contains the source term for all x, y, and z edges of the mesh.

    Notes
    -----
    You can create a closed loop by setting the first and end point to be the same.
    """
    if isinstance(mesh, discretize.TensorMesh):
        return _poly_line_source_tens(mesh, locs)
    elif isinstance(mesh, discretize.TreeMesh):
        return _poly_line_source_tree(mesh, locs)


def _poly_line_source_tens(mesh, locs):
    """
    Given a tensor product mesh with origin at (x0,y0,z0) and cell sizes
    hx, hy, hz, compute the source vector for a unit current flowing along
    the polygon with vertices px, py, pz.
    The 3-D arrays sx, sy, sz contain the source terms for all x/y/z-edges
    of the tensor product mesh.

    Modified from matlab code:

        getSourceTermLineCurrentPolygon(x0,y0,z0,hx,hy,hz,px,py,pz)
        Christoph Schwarzbach, February 2014

    """
    # Get some mesh properties
    nx, ny, nz = mesh.shape_cells
    hx, hy, hz = mesh.h
    x = mesh.nodes_x
    y = mesh.nodes_y
    z = mesh.nodes_z

    # Source points
    px = locs[:, 0]
    py = locs[:, 1]
    pz = locs[:, 2]

    # discrete edge function
    sx = np.zeros((nx, ny + 1, nz + 1))
    sy = np.zeros((nx + 1, ny, nz + 1))
    sz = np.zeros((nx + 1, ny + 1, nz))

    # number of line segments
    nP = len(px) - 1

    # check that all polygon vertices are inside the mesh
    for ip in range(nP + 1):
        ax = px[ip]
        ay = py[ip]
        az = pz[ip]
        ix = _findlast(np.logical_and(ax >= x[: nx - 1], ax <= x[1:nx]))
        iy = _findlast(np.logical_and(ay >= y[: ny - 1], ay <= y[1:ny]))
        iz = _findlast(np.logical_and(az >= z[: nz - 1], az <= z[1:nz]))

        if (ix < 0) or (iy < 0) or (iz < 0):
            msg = "Polygon vertex (%.1f, %.1f, %.1f) is outside the mesh"
            print((msg) % (ax, ay, az))

    # integrate each line segment
    for ip in range(nP):
        # start and end vertices
        ax = px[ip]
        ay = py[ip]
        az = pz[ip]
        bx = px[ip + 1]
        by = py[ip + 1]
        bz = pz[ip + 1]

        # find intersection with mesh planes
        dx = bx - ax
        dy = by - ay
        dz = bz - az
        d = np.sqrt(dx**2 + dy**2 + dz**2)

        tol = d * np.finfo(float).eps

        if abs(dx) > tol:
            tx = (x - ax) / dx
            tx = tx[np.logical_and(tx >= 0, tx <= 1)]
        else:
            tx = []

        if abs(dy) > tol:
            ty = (y - ay) / dy
            ty = ty[np.logical_and(ty >= 0, ty <= 1)]
        else:
            ty = []

        if abs(dz) > tol:
            tz = (z - az) / dz
            tz = tz[np.logical_and(tz >= 0, tz <= 1)]
        else:
            tz = []

        t = np.unique(np.r_[0.0, tx, ty, tz, 1.0])
        nq = len(t) - 1
        tc = 0.5 * (t[:nq] + t[1 : nq + 1])

        for iq in range(nq):
            cx = ax + tc[iq] * dx
            cy = ay + tc[iq] * dy
            cz = az + tc[iq] * dz

            # locate cell id

            ix = _findlast(np.logical_and(cx >= x[: nx - 1], cx <= x[1:nx]))
            iy = _findlast(np.logical_and(cy >= y[: ny - 1], cy <= y[1:ny]))
            iz = _findlast(np.logical_and(cz >= z[: nz - 1], cz <= z[1:nz]))

            # local coordinates
            hxloc = hx[ix]
            hyloc = hy[iy]
            hzloc = hz[iz]
            axloc = ax + t[iq] * dx - x[ix]
            ayloc = ay + t[iq] * dy - y[iy]
            azloc = az + t[iq] * dz - z[iz]
            bxloc = ax + t[iq + 1] * dx - x[ix]
            byloc = ay + t[iq + 1] * dy - y[iy]
            bzloc = az + t[iq + 1] * dz - z[iz]
            # integrate
            sxloc, syloc, szloc = getStraightLineCurrentIntegral(
                hxloc, hyloc, hzloc, axloc, ayloc, azloc, bxloc, byloc, bzloc
            )
            # integrate
            sx[ix, iy : iy + 2, iz : iz + 2] += np.reshape(sxloc, (2, 2), order="F")
            sy[ix : ix + 2, iy, iz : iz + 2] += np.reshape(syloc, (2, 2), order="F")
            sz[ix : ix + 2, iy : iy + 2, iz] += np.reshape(szloc, (2, 2), order="F")

    return np.r_[mkvc(sx), mkvc(sy), mkvc(sz)]


def _poly_line_source_tree(mesh, locs):
    """Calculate a source term for a line current source on a OctTreeMesh

    Given an OcTreeMesh compute the source vector for a unit current flowing
    along the polygon with vertices px, py, pz.

    Parameters
    ----------
    mesh : discretize.TreeMesh
        The OctTreeMesh (3D) for the system.
    locs : (n, 3) numpy.array
        Columns contain the x, y, and z locations of consecutive points
        along the polygonal path

    Returns
    -------
    (n_edges) numpy.ndarray
        Contains the source term for all x, y, and z edges of the OcTreeMesh.
    """

    px = locs[:, 0]
    py = locs[:, 1]
    pz = locs[:, 2]

    # discrete edge vectors
    sx = np.zeros(mesh.ntEx)
    sy = np.zeros(mesh.ntEy)
    sz = np.zeros(mesh.ntEz)

    points = np.c_[px, py, pz]
    # number of line segments
    nP = len(points) - 1
    x0 = mesh.x0
    dim = mesh.dim
    for _ in range(nP + 1):
        A = points[0]
        xF = np.array([mesh.nodes_x[-1], mesh.nodes_y[-1], mesh.nodes_z[-1]])
        if np.any(A < x0) or np.any(A > xF):
            msg = "Polygon vertex ({.1f}, {.1f}, {.1f}) is outside the mesh".format(*A)
            raise ValueError(msg)

    # Loop over each line segment
    for ip in range(nP):
        # Start and end vertices
        A = points[ip]
        B = points[ip + 1]

        # Components of vector (dx, dy, dz) along the wirepath
        ds = B - A

        # Find indices of all cells intersected by the wirepath
        srcCellIds = mesh.get_cells_along_line(A, B)
        levels = mesh.cell_levels_by_index(srcCellIds)
        if isinstance(levels, np.ndarray) and np.any(levels != levels[0]):
            warnings.warn("Warning! Line path crosses a cell level change.")

        # Starts at point A!
        p0 = A
        for cell_id in srcCellIds:
            cell = mesh[cell_id]

            x0 = cell.x0
            h = cell.h
            xF = x0 + h

            edges = cell.edges
            edges_x = edges[0:4]
            edges_y = edges[4:8]
            edges_z = edges[8:12]

            # find next intersection along path
            ts = np.ones(dim)
            for i in range(dim):
                if ds[i] > 0:
                    ts[i] = (xF[i] - A[i]) / ds[i]
                elif ds[i] < 0:
                    ts[i] = (x0[i] - A[i]) / ds[i]
                else:
                    ts[i] = np.inf
            t = min(*ts, 1)  # the last value should be 1
            p1 = A + t * ds  # the next intersection point

            cA = p0 - x0
            cB = p1 - x0

            cell_s = getStraightLineCurrentIntegral(*h, *cA, *cB)

            sx[edges_x] += cell_s[0]
            sy[edges_y] += cell_s[1]
            sz[edges_z] += cell_s[2]

            p0 = p1
    s = np.r_[sx, sy, sz]
    R = mesh._deflate_edges()
    s = R.T.dot(s)

    return s


def line_through_faces(
    mesh,
    locations,
    normalize_by_area=True,
    check_divergence=True,
    tolerance_divergence=1e-9,
):
    """Define line current through cell faces

    Define the current through cell faces given path locations. Note that this
    will perform best of your path locations are at cell centers. Only paths
    that align with the mesh (e.g. a straight line in x, y, z) are currently
    supported

    Parameters
    ----------
    mesh : discretize.TreeMesh
        The OctTreeMesh (3D) for the system.
    normalize_by_area : bool, default: ``True``
        If ``True``, normalize by face area
    check_divergence : bool, default: ``True``
        If ``True``, carry out divergence check
    tolerance_divergence: float, default: 1e-9
        Tolerance for divergence check

    Returns
    -------
    (mesh.n_faces) numpy.ndarray
        Line current source on faces
    """

    current = np.zeros(mesh.n_faces)

    def not_aligned_error(i):
        raise NotImplementedError(
            "With the current implementation, the line between points "
            "must align with the axes of the mesh. The points "
            f"{locations[i, :]} and {locations[i+1, :]} do not."
        )

    # pre-processing step: find closest cell centers
    cell_centers = discretize.utils.closest_points_index(mesh, locations, "CC")
    locations = mesh.gridCC[cell_centers, :]

    # next step: find segments between lines
    for i in range(locations.shape[0] - 1):
        dimension = np.nonzero(np.abs(locations[i, :] - locations[i + 1, :]))[0]
        if len(dimension) > 1:
            not_aligned_error(i)
        dimension = dimension[0]
        direction = np.sign(locations[i, dimension] - locations[i + 1, dimension])

        if dimension == 0:
            grid_loc = "x"
            current_inds = slice(0, mesh.n_faces_x)
        elif dimension == 1:
            grid_loc = "y"
            start = mesh.n_faces_x
            current_inds = slice(start, start + mesh.n_faces_y)
        elif dimension == 2:
            grid_loc = "z"
            start = mesh.n_faces_x + mesh.n_faces_y
            current_inds = slice(start, start + mesh.n_faces_z)

        # interpolate to closest face
        loca = discretize.utils.closest_points_index(
            mesh, locations[i, :], f"F{grid_loc}"
        )
        locb = discretize.utils.closest_points_index(
            mesh, locations[i + 1, :], f"F{grid_loc}"
        )

        if len(loca) > 1 or len(locb) > 1:
            raise Exception(
                "Current across multiple faces is not implemented. "
                "Please put path through a cell rather than along edges"
            )

        grid = getattr(mesh, f"faces_{grid_loc}")
        loca = grid[loca[0]]
        locb = grid[locb[0]]

        # find all faces between these points
        if dimension == 0:
            xlocs = np.r_[locations[i, 0], locations[i + 1, 0]]

            if not (np.allclose(loca[1], locb[1]) and np.allclose(loca[2], locb[2])):
                not_aligned_error(i)

            ylocs = loca[1] + mesh.h[1].min() / 4 * np.r_[-1, 1]
            zlocs = loca[2] + mesh.h[2].min() / 4 * np.r_[-1, 1]

        elif dimension == 1:
            ylocs = np.r_[locations[i, 1], locations[i + 1, 1]]

            if not (np.allclose(loca[0], locb[0]) and np.allclose(loca[2], locb[2])):
                not_aligned_error(i)

            xlocs = loca[0] + mesh.h[0].min() / 4 * np.r_[-1, 1]
            zlocs = loca[2] + mesh.h[2].min() / 4 * np.r_[-1, 1]

        elif dimension == 2:
            zlocs = np.r_[locations[i, 2], locations[i + 1, 2]]

            if not (np.allclose(loca[0], locb[0]) and np.allclose(loca[1], locb[1])):
                not_aligned_error(i)

            xlocs = loca[0] + mesh.h[0].min() / 4 * np.r_[-1, 1]
            ylocs = loca[1] + mesh.h[1].min() / 4 * np.r_[-1, 1]

        src_inds = (
            (grid[:, 0] >= xlocs.min())
            & (grid[:, 0] <= xlocs.max())
            & (grid[:, 1] >= ylocs.min())
            & (grid[:, 1] <= ylocs.max())
            & (grid[:, 2] >= zlocs.min())
            & (grid[:, 2] <= zlocs.max())
        )

        current[current_inds][src_inds] = direction

    if normalize_by_area:
        current = current / mesh.face_areas

    # check that there is only a divergence at the ends if not a loop
    if check_divergence:
        div = mesh.cell_volumes * mesh.face_divergence * current
        nonzero = np.abs(div) > tolerance_divergence

        # check if the source is a loop or grounded
        if not np.allclose(locations[0, :], locations[-1, :]):  # grounded source
            if nonzero.sum() > 2:
                raise Exception(
                    "The source path is not connected. Check that all points go through cell centers"
                )
            if np.abs(div.sum()) > tolerance_divergence:
                raise Exception(
                    "The integral of the divergence is not zero. Something is wrong :("
                )
        else:  # loop source
            if nonzero.sum() > 0:
                raise Exception(
                    "The source path is not connected. Check that all points go through cell centers"
                )

    return current


def getSourceTermLineCurrentPolygon(xorig, hx, hy, hz, px, py, pz):
    """getSourceTermLineCurrentPolygon is deprecated. Use :func:`segmented_line_current_source_term`"""
    raise NotImplementedError(
        "getSourceTermLineCurrentPolygon has been deprecated and will be"
        "removed in SimPEG 0.17.0. Please use segmented_line_current_source_term.",
    )



"""
=================================================================
Base EM (:mod:`SimPEG.electromagnetics`)
=================================================================
.. currentmodule:: SimPEG.electromagnetics

About ``electromagnetics``

Base Classes
============
.. autosummary::
  :toctree: generated/

  base.BaseEMSimulation
  base.BaseEMSrc

Analytic Functions
==================
Many of these functions call functions from `geoana`, and we recommend calling
those functions directly from that package.

.. autosummary::
  :toctree: generated/

  analytics.h[2]AnalyticDipoleT
  analytics.h[2]AnalyticCentLoopT
  analytics.h[2]AnalyticDipoleF
  analytics.getCasingEphiMagDipole
  analytics.getCasingHrMagDipole
  analytics.getCasingHzMagDipole
  analytics.getCasingBrMagDipole
  analytics.getCasingBzMagDipole

"""
from scipy.constants import mu_0, epsilon_0

from . import time_domain
from . import frequency_domain
from . import natural_source
from . import analytics
from . import utils
from .static import resistivity, induced_polarization, spectral_induced_polarization



r"""
=============================================
Regularization (:mod:`SimPEG.regularization`)
=============================================
.. currentmodule:: SimPEG.regularization

If there is one model that has a misfit that equals the desired tolerance,
then there are infinitely many other models which can fit to the same degree.
The challenge is to find that model which has the desired characteristics and
is compatible with a priori information. A single model can be selected from
an infinite ensemble by measuring the length, or norm, of each model. Then a
smallest, or sometimes largest, member can be isolated. Our goal is to design
a norm that embodies our prior knowledge and, when minimized, yields a
realistic candidate for the solution of our problem. The norm can penalize
variation from a reference model, spatial derivatives of the model, or some
combination of these.

WeightedLeastSquares Regularization
===================================

Here we will define regularization of a model, m, in general however, this
should be thought of as (m-m_ref) but otherwise it is exactly the same:

.. math::

    R(m) =
        \int_\Omega
        \frac{\alpha_x}{2}
            \left( \frac{\partial m}{\partial x} \right)^2
        + \frac{\alpha_y}{2}
            \left( \frac{\partial m}{\partial y} \right)^2
        \partial v

Our discrete gradient operator works on cell centers and gives the derivative
on the cell faces, which is not where we want to be evaluating this integral.
We need to average the values back to the cell-centers before we integrate. To
avoid null spaces, we square first and then average. In 2D with ij notation it
looks like this:

.. math::

    R(m) \approx
        \sum_{ij}
        \left[
            \frac{\alpha_x}{2} \left[
                \left( \frac{m_{i+1,j} - m_{i,j}}{h} \right)^2
                + \left( \frac{m_{i,j} - m_{i-1,j}}{h} \right)^2
            \right] \\
            + \frac{\alpha_y}{2} \left[
                \left( \frac{m_{i,j+1} - m_{i,j}}{h} \right)^2
                + \left( \frac{m_{i,j} - m_{i,j-1}}{h} \right)^2
            \right]
        \right] h^2

If we let D_1 be the derivative matrix in the x direction

.. math::

    \mathbf{D}_1 = \mathbf{I}_2\otimes\mathbf{d}_1

.. math::

    \mathbf{D}_2 = \mathbf{d}_2\otimes\mathbf{I}_1

Where d_1 is the one dimensional derivative:

.. math::

    \mathbf{d}_1 = \frac{1}{h} \left[ \begin{array}{cccc}
    -1 & 1 & & \\
     & \ddots & \ddots&\\
     &  & -1 & 1\end{array} \right]

.. math::

    R(m) \approx
        \mathbf{v}^\top \left[
            \frac{\alpha_x}{2} \mathbf{A}_1 (\mathbf{D}_1 m) \odot (\mathbf{D}_1 m)
            + \frac{\alpha_y}{2} \mathbf{A}_2 (\mathbf{D}_2 m) \odot (\mathbf{D}_2 m)
        \right]

Recall that this is really a just point wise multiplication, or a diagonal
matrix times a vector. When we multiply by something in a diagonal we can
interchange and it gives the same results (i.e. it is point wise)

.. math::

    \mathbf{a\odot b}
        = \text{diag}(\mathbf{a})\mathbf{b}
        = \text{diag}(\mathbf{b})\mathbf{a}
        = \mathbf{b\odot a}

and the transpose also is true (but the sizes have to make sense...):

.. math::

    \mathbf{a}^\top\text{diag}(\mathbf{b}) = \mathbf{b}^\top\text{diag}(\mathbf{a})

So R(m) can simplify to:

.. math::

    R(m) \approx
        \mathbf{m}^\top \left[
            \frac{\alpha_x}{2} \mathbf{D}_1^\top
            \text{diag} (\mathbf{A}_1^\top \mathbf{v}) \mathbf{D}_1
            + \frac{\alpha_y}{2} \mathbf{D}_2^\top
            \text{diag} (\mathbf{A}_2^\top \mathbf{v}) \mathbf{D}_2
        \right]
        \mathbf{m}

We will define W_x as:

.. math::

    \mathbf{W}_x =
        \sqrt{\alpha_x}
        \text{diag} \left(\sqrt{\mathbf{A}_1^\top\mathbf{v}}\right) \mathbf{D}_1


And then W as a tall matrix of all of the different regularization terms:

.. math::

    \mathbf{W} = \left[ \begin{array}{c}
    \mathbf{W}_s\\
    \mathbf{W}_x\\
    \mathbf{W}_y\end{array} \right]

Then we can write

.. math::

    R(m) \approx \frac{1}{2}\mathbf{m^\top W^\top W m}


The API
=======

Least Squares Regularizations
-----------------------------
.. autosummary::
  :toctree: generated/

  WeightedLeastSquares
  Smallness
  SmoothnessFirstOrder
  SmoothnessSecondOrder

Sparse Regularizations
----------------------
We have also implemented several sparse regularizations with a variable norm.

.. autosummary::
  :toctree: generated/

  Sparse
  SparseSmallness
  SparseSmoothness

Joint Regularizations
---------------------
There are several joint inversion regularizers available

.. autosummary::
  :toctree: generated/

  CrossGradient
  JointTotalVariation
  PGI
  PGIsmallness
  LinearCorrespondence

Base Regularization classes
---------------------------
.. autosummary::
  :toctree: generated/

  RegularizationMesh
  BaseRegularization
  BaseSimilarityMeasure
  BaseSparse

"""
from ..utils.code_utils import deprecate_class
from .base import (
    BaseRegularization,
    WeightedLeastSquares,
    BaseSimilarityMeasure,
    Smallness,
    SmoothnessFirstOrder,
    SmoothnessSecondOrder,
)
from .regularization_mesh import RegularizationMesh
from .sparse import BaseSparse, SparseSmallness, SparseSmoothness, Sparse
from .pgi import PGIsmallness, PGI
from .cross_gradient import CrossGradient
from .correspondence import LinearCorrespondence
from .jtv import JointTotalVariation


@deprecate_class(removal_version="0.19.0", future_warn=True)
class SimpleSmall(Smallness):
    """Deprecated class, replaced by Smallness."""

    pass


@deprecate_class(removal_version="0.19.0", future_warn=True)
class SimpleSmoothDeriv(SmoothnessFirstOrder):
    """Deprecated class, replaced by SmoothnessFirstOrder."""

    pass


@deprecate_class(removal_version="0.19.0", future_warn=True)
class Simple(WeightedLeastSquares):
    """Deprecated class, replaced by WeightedLeastSquares."""

    def __init__(self, mesh=None, alpha_x=1.0, alpha_y=1.0, alpha_z=1.0, **kwargs):
        # These alphas are now refered to as length_scalse in the
        # new WeightedLeastSquares regularization
        super().__init__(
            mesh=mesh,
            length_scale_x=alpha_x,
            length_scale_y=alpha_y,
            length_scale_z=alpha_z,
            **kwargs
        )


@deprecate_class(removal_version="0.19.0", future_warn=True)
class Tikhonov(WeightedLeastSquares):
    """Deprecated class, replaced by WeightedLeastSquares."""

    def __init__(
        self, mesh=None, alpha_s=1e-6, alpha_x=1.0, alpha_y=1.0, alpha_z=1.0, **kwargs
    ):
        super().__init__(
            mesh=mesh,
            alpha_s=alpha_s,
            alpha_x=alpha_x,
            alpha_y=alpha_y,
            alpha_z=alpha_z,
            **kwargs
        )


@deprecate_class(removal_version="0.19.0", future_warn=True)
class Small(Smallness):
    """Deprecated class, replaced by Smallness."""

    pass


@deprecate_class(removal_version="0.19.0", future_warn=True)
class SmoothDeriv(SmoothnessFirstOrder):
    """Deprecated class, replaced by SmoothnessFirstOrder."""

    pass


@deprecate_class(removal_version="0.19.0", future_warn=True)
class SmoothDeriv2(SmoothnessSecondOrder):
    """Deprecated class, replaced by SmoothnessSecondOrder."""

    pass


@deprecate_class(removal_version="0.19.0", future_warn=True)
class PGIwithNonlinearRelationshipsSmallness(PGIsmallness):
    """Deprecated class, replaced by PGIsmallness."""

    def __init__(self, gmm, **kwargs):
        super().__init__(gmm, non_linear_relationships=True, **kwargs)


@deprecate_class(removal_version="0.19.0", future_warn=True)
class PGIwithRelationships(PGI):
    """Deprecated class, replaced by PGI."""

    def __init__(self, mesh, gmmref, **kwargs):
        super().__init__(mesh, gmmref, non_linear_relationships=True, **kwargs)



import numpy as np
import sys
import gc

from .... import props
from .data import Data
from ....utils import sdiag, validate_type, validate_active_indices
import scipy.sparse as sp

from ..induced_polarization.simulation import BaseIPSimulation
from ..induced_polarization import (
    Simulation3DCellCentered as BaseSimulation3DCellCentered,
)
from ..induced_polarization import Simulation3DNodal as BaseSimulation3DNodal
from .survey import Survey


class BaseSIPSimulation(BaseIPSimulation):
    tau, tauMap, tauDeriv = props.Invertible("Time constant (s)")
    taui, tauiMap, tauiDeriv = props.Invertible("Inverse of time constant (1/s)")
    props.Reciprocal(tau, taui)

    c, cMap, cDeriv = props.Invertible("Frequency dependency")

    Ainv = None
    _f = None
    _Jmatrix = None
    _dc_voltage_set = False

    _eta_store = None
    _taui_store = None
    _c_store = None
    _pred = None

    def __init__(
        self,
        mesh,
        survey=None,
        tau=0.1,
        tauMap=None,
        taui=None,
        tauiMap=None,
        c=0.5,
        cMap=None,
        storeJ=False,
        actinds=None,
        storeInnerProduct=True,
        **kwargs
    ):
        super().__init__(mesh=mesh, survey=survey, **kwargs)
        self.tau = tau
        self.taui = taui
        self.tauMap = tauMap
        self.tauiMap = tauiMap
        self.c = c
        self.cMap = cMap
        self.storeJ = storeJ
        self.storeInnerProduct = storeInnerProduct
        self.actinds = actinds

    @property
    def survey(self):
        """The SIP survey object.

        Returns
        -------
        SimPEG.electromagnetics.static.spectral_induced_polarization.survey.Survey
        """
        if self._survey is None:
            raise AttributeError("Simulation must have a survey")
        return self._survey

    @survey.setter
    def survey(self, value):
        if value is not None:
            value = validate_type("survey", value, Survey, cast=False)
        self._survey = value

    @property
    def actinds(self):
        """Active indices when storing J.

        Returns
        -------
        (mesh.n_cells) numpy.ndarray of bool
        """
        return self._actinds

    @actinds.setter
    def actinds(self, value):
        if self.storeJ:
            if value is None:
                value = np.ones(self.mesh.n_cells, dtype=bool)
            value = validate_active_indices("actinds", value, self.mesh.n_cells)
            self._actinds = value
            self._P = sp.eye(self.mesh.n_cells, format="csr")[:, value]

    @property
    def storeInnerProduct(self):
        """Whether to store inner product matrices

        Returns
        -------
        bool
        """
        return self._storeInnerProduct

    @storeInnerProduct.setter
    def storeInnerProduct(self, value):
        self._storeInnerProduct = validate_type("storeInnerProduct", value, bool)

    @property
    def storeJ(self):
        """Whether to store the sensitivity matrix

        Returns
        -------
        bool
        """
        return self._storeJ

    @storeJ.setter
    def storeJ(self, value):
        self._storeJ = validate_type("storeJ", value, bool)

    @property
    def n(self):
        return self.mesh.n_nodes

    @property
    def sigmaDeriv(self):
        if self.storeJ:
            dsigma_dlogsigma = sdiag(self.sigma) * self._P
        else:
            dsigma_dlogsigma = sdiag(self.sigma)
        return -dsigma_dlogsigma

    @property
    def rhoDeriv(self):
        if self.storeJ:
            drho_dlogrho = sdiag(self.rho) * self._P
        else:
            drho_dlogrho = sdiag(self.rho)
        return drho_dlogrho

    @property
    def etaDeriv_store(self):
        if getattr(self, "_etaDeriv_store", None) is None:
            self._etaDeriv_store = self.etaDeriv
        return self._etaDeriv_store

    @property
    def tauiDeriv_store(self):
        if getattr(self, "_tauiDeriv_store", None) is None:
            self._tauiDeriv_store = self.tauiDeriv
        return self._tauiDeriv_store

    @property
    def tauDeriv_store(self):
        if getattr(self, "_tauDeriv_store", None) is None:
            self._tauDeriv_store = self.tauDeriv
        return self._tauDeriv_store

    @property
    def cDeriv_store(self):
        if getattr(self, "_cDeriv_store", None) is None:
            self._cDeriv_store = self.cDeriv
        return self._cDeriv_store

    def get_t_over_tau(self, t):
        taui = self._taui_store
        return t * taui

    def get_exponent(self, t):
        c = self._c_store
        t_over_tau = self.get_t_over_tau(t)
        return (t_over_tau) ** c

    def get_peta_step_off(self, exponent):
        eta = self._eta_store
        peta = eta * np.exp(-exponent)
        return peta

    def get_peta_pulse_off(self, t):
        """
        Compute pseudo-chargeability from a single pulse waveform
        """
        T = self.survey.T
        exponent_0 = self.get_exponent(t)
        exponent_1 = self.get_exponent(t + T / 4.0)
        peta = self.get_peta_step_off(exponent_0) - self.get_peta_step_off(exponent_1)
        return peta

    def get_multi_pulse_response(self, t, pulse_func):
        n_pulse = self.survey.n_pulse
        T = self.survey.T
        peta = np.zeros(self._eta_store.shape[0], float, order="C")
        for i_pulse in range(n_pulse):
            factor = (-1) ** i_pulse * (n_pulse - i_pulse)
            peta += pulse_func(t + T / 2 * i_pulse) * factor
        return peta / n_pulse

    def get_peta(self, t):
        n_pulse = self.survey.n_pulse
        if n_pulse == 0:
            exponent = self.get_exponent(t)
            return self.get_peta_step_off(exponent)
        else:
            return self.get_multi_pulse_response(t, self.get_peta_pulse_off)

    def get_peta_eta_deriv_step_off(self, exponent):
        return np.exp(-exponent)

    def get_peta_eta_deriv_pulse_off(self, t):
        """
        Compute derivative of pseudo-chargeability w.r.t eta from a single pulse waveform
        """
        T = self.survey.T
        exponent_0 = self.get_exponent(t)
        exponent_1 = self.get_exponent(t + T / 4.0)
        peta_eta_deriv = self.get_peta_eta_deriv_step_off(
            exponent_0
        ) - self.get_peta_eta_deriv_step_off(exponent_1)
        return peta_eta_deriv

    def get_peta_eta_deriv(self, t):
        n_pulse = self.survey.n_pulse
        if n_pulse == 0:
            exponent = self.get_exponent(t)
            return self.get_peta_eta_deriv_step_off(exponent)
        else:
            return self.get_multi_pulse_response(t, self.get_peta_eta_deriv_pulse_off)

    def PetaEtaDeriv(self, t, v, adjoint=False):
        etaDeriv = self.etaDeriv_store

        v = np.array(v, dtype=float)
        dpetadeta = self.get_peta_eta_deriv(t)

        if adjoint:
            if v.ndim == 1:
                return etaDeriv.T * (dpetadeta * v)
            else:
                return etaDeriv.T * (sdiag(dpetadeta) * v)
        else:
            return dpetadeta * (etaDeriv * v)

    def get_peta_taui_deriv_step_off(self, exponent):
        eta = self._eta_store
        taui = self._taui_store
        c = self._c_store
        peta_taui_deriv = -c * eta / taui * exponent * np.exp(-exponent)
        return peta_taui_deriv

    def get_peta_taui_deriv_pulse_off(self, t):
        """
        Compute derivative of pseudo-chargeability w.r.t eta from a single pulse waveform
        """
        T = self.survey.T
        exponent_0 = self.get_exponent(t)
        exponent_1 = self.get_exponent(t + T / 4.0)
        peta_taui_deriv = self.get_peta_taui_deriv_step_off(
            exponent_0
        ) - self.get_peta_taui_deriv_step_off(exponent_1)
        return peta_taui_deriv

    def get_peta_taui_deriv(self, t):
        n_pulse = self.survey.n_pulse
        if n_pulse == 0:
            exponent = self.get_exponent(t)
            return self.get_peta_taui_deriv_step_off(exponent)
        else:
            return self.get_multi_pulse_response(t, self.get_peta_taui_deriv_pulse_off)

    def PetaTauiDeriv(self, t, v, adjoint=False):
        v = np.array(v, dtype=float)
        tauiDeriv = self.tauiDeriv_store
        dpetadtaui = self.get_peta_taui_deriv(t)

        if adjoint:
            if v.ndim == 1:
                return tauiDeriv.T * (dpetadtaui * v)
            else:
                return tauiDeriv.T * (sdiag(dpetadtaui) * v)
        else:
            return dpetadtaui * (tauiDeriv * v)

    def get_peta_c_deriv_step_off(self, exponent, t_over_tau):
        eta = self._eta_store
        peta_c_deriv = -eta * (exponent) * np.exp(-exponent) * np.log(t_over_tau)
        return peta_c_deriv

    def get_peta_c_deriv_pulse_off(self, t):
        """
        Compute derivative of pseudo-chargeability w.r.t eta from a single pulse waveform
        """
        T = self.survey.T
        exponent_0 = self.get_exponent(t)
        exponent_1 = self.get_exponent(t + T / 4.0)
        t_over_tau_0 = self.get_t_over_tau(t)
        t_over_tau_1 = self.get_t_over_tau(t + T / 4.0)

        peta_c_deriv = self.get_peta_c_deriv_step_off(
            exponent_0, t_over_tau_0
        ) - self.get_peta_c_deriv_step_off(exponent_1, t_over_tau_1)
        return peta_c_deriv

    def get_peta_c_deriv(self, t):
        n_pulse = self.survey.n_pulse
        if n_pulse == 0:
            exponent = self.get_exponent(t)
            t_over_tau = self.get_t_over_tau(t)
            return self.get_peta_c_deriv_step_off(exponent, t_over_tau)
        else:
            return self.get_multi_pulse_response(t, self.get_peta_c_deriv_pulse_off)

    def PetaCDeriv(self, t, v, adjoint=False):
        v = np.array(v, dtype=float)
        cDeriv = self.cDeriv_store
        dpetadc = self.get_peta_c_deriv(t)
        if adjoint:
            if v.ndim == 1:
                return cDeriv.T * (dpetadc * v)
            else:
                return cDeriv.T * (sdiag(dpetadc) * v)
        else:
            return dpetadc * (cDeriv * v)

    def fields(self, m):
        if self.verbose:
            print(">> Compute DC fields")
        if self._f is None:
            # re-uses the DC simulation's fields method
            # This grabs the class just above BasIPSimulation
            self._f = super(BaseIPSimulation, self).fields(None)

        if not self._dc_voltage_set:
            try:
                f = self.fields_to_space(self._f)
            except AttributeError:
                f = self._f
            # loop through receievers to check if they need to set the _dc_voltage
            for src in self.survey.source_list:
                for rx in src.receiver_list:
                    if rx.data_type == "apparent_chargeability":
                        rx.data_type = "volt"
                        rx._dc_voltage = rx.eval(src, self.mesh, f)
                        if rx.storeProjections:
                            P = rx._Ps[self.mesh]
                            rx._Ps[self.mesh] = sdiag(1.0 / rx.dc_voltage) * P
                        rx.data_type = "apparent_chargeability"

        self._pred = self.forward(m, f=self._f)

        return self._f

    # @profile
    def getJ(self, m, f=None):
        """
        Generate Full sensitivity matrix
        """

        if self._Jmatrix is not None:
            return self._Jmatrix
        else:
            if self.verbose:
                print("Calculating J and storing")
            if f is None:
                f = self.fields(m)

            Jt = np.zeros(
                (self._P.shape[1], int(self.survey.nD / self.survey.unique_times.size)),
                order="F",
            )
            istrt = int(0)
            iend = int(0)

            for isrc, src in enumerate(self.survey.source_list):
                if self.verbose:
                    sys.stdout.write(("\r %d / %d") % (isrc + 1, self.survey.nSrc))
                    sys.stdout.flush()
                u_src = f[src, self._solutionType]
                for rx in src.receiver_list:
                    P = rx.getP(self.mesh, rx.projGLoc(f)).toarray()
                    ATinvdf_duT = self.Ainv * (P.T)
                    dA_dmT = self.getADeriv(u_src, ATinvdf_duT, adjoint=True)
                    iend = istrt + rx.nD
                    if rx.nD == 1:
                        Jt[:, istrt] = -dA_dmT
                    else:
                        Jt[:, istrt:iend] = -dA_dmT
                    istrt += rx.nD

            self._Jmatrix = Jt.T
            collected = gc.collect()
            if self.verbose:
                collected = gc.collect()
                print("Garbage collector: collected %d objects." % (collected))
            # clean all factorization
            if self.Ainv is not None:
                self.Ainv.clean()

            return self._Jmatrix

    def getJtJdiag(self, m, Wd, f=None):
        """
        Compute JtJ using adjoint problem. Still we never form
        JtJ
        """
        if self.verbose:
            print(">> Compute trace(JtJ)")
        ntime = len(self.survey.unique_times)
        JtJdiag = np.zeros_like(m)
        J = self.getJ(m, f=f)
        wd = Wd.diagonal().reshape((self.survey.locations_n, ntime), order="F")
        for tind in range(ntime):
            t = self.survey.unique_times[tind]
            Jtv = self._P * J.T * sdiag(wd[:, tind])
            JtJdiag += (
                (self.PetaEtaDeriv(t, Jtv, adjoint=True) ** 2).sum(axis=1)
                + (self.PetaTauiDeriv(t, Jtv, adjoint=True) ** 2).sum(axis=1)
                + (self.PetaCDeriv(t, Jtv, adjoint=True) ** 2).sum(axis=1)
            )
        return JtJdiag

    # @profile
    def forward(self, m, f=None):
        if self.verbose:
            print(">> Compute predicted data")

        self.model = m

        self._eta_store = self.eta
        self._taui_store = self.taui
        self._c_store = self.c

        Jv = []

        # When sensitivity matrix is stored
        if self.storeJ:
            J = self.getJ(m, f=f)

            ntime = len(self.survey.unique_times)

            self.model = m
            for tind in range(ntime):
                Jv.append(
                    J.dot(self._P.T * self.get_peta(self.survey.unique_times[tind]))
                )
            return np.hstack(Jv)

        # Do not store sensitivity matrix (memory-wise efficient)
        else:
            if f is None:
                f = self.fields(m)

            # A = self.getA()
            for tind in range(len(self.survey.unique_times)):
                # Pseudo-chareability
                t = self.survey.unique_times[tind]
                v = self.get_peta(t)
                for src in self.survey.source_list:
                    u_src = f[src, self._solutionType]  # solution vector
                    dA_dm_v = self.getADeriv(u_src, v)
                    dRHS_dm_v = self.getRHSDeriv(src, v)
                    du_dm_v = self.Ainv * (-dA_dm_v + dRHS_dm_v)
                    for rx in src.receiver_list:
                        timeindex = rx.getTimeP(self.survey.unique_times)
                        if timeindex[tind]:
                            df_dmFun = getattr(
                                f, "_{0!s}Deriv".format(rx.projField), None
                            )
                            df_dm_v = df_dmFun(src, du_dm_v, v, adjoint=False)
                            Jv.append(rx.evalDeriv(src, self.mesh, f, df_dm_v))

            return np.hstack(Jv)

    def dpred(self, m, f=None):
        r"""
        Predicted data.

        .. math::

            d_\text{pred} = Pf(m)

        """
        if f is None:
            f = self.fields(m)
        return self._pred
        # return self.forward(m, f=f)

    def Jvec(self, m, v, f=None):
        self.model = m

        Jv = []

        # When sensitivity matrix is stored
        if self.storeJ:
            J = self.getJ(m, f=f)
            ntime = len(self.survey.unique_times)

            for tind in range(ntime):
                t = self.survey.unique_times[tind]
                v0 = self.PetaEtaDeriv(t, v)
                v1 = self.PetaTauiDeriv(t, v)
                v2 = self.PetaCDeriv(t, v)
                PTv = self._P.T * (v0 + v1 + v2)
                Jv.append(J.dot(PTv))

            return np.hstack(Jv)

        # Do not store sensitivity matrix (memory-wise efficient)
        else:
            if f is None:
                f = self.fields(m)

            for tind in range(len(self.survey.unique_times)):
                t = self.survey.unique_times[tind]
                v0 = self.PetaEtaDeriv(t, v)
                v1 = self.PetaTauiDeriv(t, v)
                v2 = self.PetaCDeriv(t, v)

                for src in self.survey.source_list:
                    u_src = f[src, self._solutionType]  # solution vector
                    dA_dm_v = self.getADeriv(u_src, v0 + v1 + v2)
                    dRHS_dm_v = self.getRHSDeriv(src, v0 + v1 + v2)
                    du_dm_v = self.Ainv * (-dA_dm_v + dRHS_dm_v)

                    for rx in src.receiver_list:
                        # Assume same # of time
                        # timeindex = rx.getTimeP(self.survey.unique_times)
                        # if timeindex[tind]:
                        Jv_temp = rx.evalDeriv(src, self.mesh, f, du_dm_v)
                        Jv.append(Jv_temp)

            return np.hstack(Jv)

    def Jtvec(self, m, v, f=None):
        self.model = m

        # When sensitivity matrix is stored
        if self.storeJ:
            J = self.getJ(m, f=f)
            ntime = len(self.survey.unique_times)
            Jtvec = np.zeros(m.size)
            v = v.reshape((int(self.survey.nD / ntime), ntime), order="F")

            for tind in range(ntime):
                t = self.survey.unique_times[tind]
                Jtv = self._P * J.T.dot(v[:, tind])
                Jtvec += (
                    self.PetaEtaDeriv(t, Jtv, adjoint=True)
                    + self.PetaTauiDeriv(t, Jtv, adjoint=True)
                    + self.PetaCDeriv(t, Jtv, adjoint=True)
                )

            return Jtvec

        # Do not store sensitivity matrix (memory-wise efficient)
        else:
            if f is None:
                f = self.fields(m)

            # Ensure v is a data object.
            if not isinstance(v, Data):
                v = Data(self.survey, v)

            Jtv = np.zeros(m.size, dtype=float)
            n_time = len(self.survey.unique_times)
            du_dmT = np.zeros((self.mesh.nC, n_time), dtype=float, order="F")

            for tind in range(n_time):
                t = self.survey.unique_times[tind]
                for src in self.survey.source_list:
                    u_src = f[src, self._solutionType]
                    for rx in src.receiver_list:
                        # Ignore case when each rx has different # of times

                        # timeindex = rx.getTimeP(self.survey.unique_times)
                        # if timeindex[tind]:
                        # wrt f, need possibility wrt m
                        PTv = rx.evalDeriv(
                            src, self.mesh, f, v[src, rx, t], adjoint=True
                        )
                        df_duTFun = getattr(f, "_{0!s}Deriv".format(rx.projField), None)
                        df_duT, _ = df_duTFun(src, None, PTv, adjoint=True)
                        ATinvdf_duT = self.Ainv * df_duT
                        du_dmT[:, tind] += -self.getADeriv(
                            u_src, ATinvdf_duT, adjoint=True
                        )

                Jtv += (
                    self.PetaEtaDeriv(
                        self.survey.unique_times[tind], du_dmT[:, tind], adjoint=True
                    )
                    + self.PetaTauiDeriv(
                        self.survey.unique_times[tind], du_dmT[:, tind], adjoint=True
                    )
                    + self.PetaCDeriv(
                        self.survey.unique_times[tind], du_dmT[:, tind], adjoint=True
                    )
                )
            return Jtv

    @property
    def deleteTheseOnModelUpdate(self):
        toDelete = [
            "_etaDeriv_store",
            "_tauiDeriv_store",
            "_cDeriv_store",
            "_tauDeriv_store",
        ]
        return toDelete


class Simulation3DCellCentered(BaseSIPSimulation, BaseSimulation3DCellCentered):
    """
    3D cell centered Spectral IP problem
    """


class Simulation3DNodal(BaseSIPSimulation, BaseSimulation3DNodal):
    """
    3D nodal Spectral IP problem
    """


Simulation3DCellCentred = Simulation3DCellCentered



from scipy.constants import mu_0
import numpy as np
from scipy import sparse as sp
from scipy.special import roots_legendre
from empymod.transform import get_dlf_points

from ..simulation import BaseSimulation

# from .time_domain.sources import MagDipole as t_MagDipole, CircularLoop as t_CircularLoop
# from .frequency_domain.sources import MagDipole as f_MagDipole, CircularLoop as f_CircularLoop

from .. import utils
from ..utils import (
    validate_string,
    validate_ndarray_with_shape,
    validate_type,
    validate_integer,
)
from .. import props
from empymod.utils import check_hankel

__all__ = ["BaseEM1DSimulation"]

###############################################################################
#                                                                             #
#                             Base EM1D Simulation                            #
#                                                                             #
###############################################################################


class BaseEM1DSimulation(BaseSimulation):
    """
    Base simulation class for simulating the EM response over a 1D layered Earth
    for a single sounding. The simulation computes the fields by solving the
    Hankel transform solutions from Electromagnetic Theory for Geophysical
    Applications: Chapter 4 (Ward and Hohmann, 1988).
    """

    _formulation = "1D"
    _coefficients_set = False

    # Properties for electrical conductivity/resistivity
    sigma, sigmaMap, sigmaDeriv = props.Invertible(
        "Electrical conductivity at infinite frequency (S/m)"
    )
    rho, rhoMap, rhoDeriv = props.Invertible("Electrical resistivity (Ohm m)")
    props.Reciprocal(sigma, rho)

    eta = props.PhysicalProperty("Intrinsic chargeability (V/V), 0 <= eta < 1")
    tau = props.PhysicalProperty("Time constant for Cole-Cole model (s)")
    c = props.PhysicalProperty("Frequency Dependency for Cole-Cole model, 0 < c < 1")

    # Properties for magnetic susceptibility
    mu, muMap, muDeriv = props.Invertible(
        "Magnetic permeability at infinite frequency (SI)"
    )
    dchi = props.PhysicalProperty(
        "DC magnetic susceptibility for viscous remanent magnetization contribution (SI)"
    )
    tau1 = props.PhysicalProperty(
        "Lower bound for log-uniform distribution of time-relaxation constants for viscous remanent magnetization (s)"
    )
    tau2 = props.PhysicalProperty(
        "Upper bound for log-uniform distribution of time-relaxation constants for viscous remanent magnetization (s)"
    )

    # Additional properties
    h, hMap, hDeriv = props.Invertible("Receiver Height (m), h > 0")

    thicknesses, thicknessesMap, thicknessesDeriv = props.Invertible(
        "layer thicknesses (m)"
    )

    def __init__(
        self,
        sigma=None,
        sigmaMap=None,
        rho=None,
        rhoMap=None,
        thicknesses=None,
        thicknessesMap=None,
        mu=mu_0,
        muMap=None,
        h=None,
        hMap=None,
        eta=0.0,
        tau=1.0,
        c=0.5,
        dchi=0.0,
        tau1=1.0e-10,
        tau2=10.0,
        hankel_filter="key_101_2009",
        fix_Jmatrix=False,
        topo=None,
        n_points_per_path=3,
        **kwargs,
    ):
        super().__init__(mesh=None, **kwargs)
        self.sigma = sigma
        self.rho = rho
        self.sigmaMap = sigmaMap
        self.rhoMap = rhoMap
        self.mu = mu
        self.muMap = muMap
        self.h = h
        self.hMap = hMap
        if thicknesses is None:
            thicknesses = np.array([])
        self.thicknesses = thicknesses
        self.thicknessesMap = thicknessesMap
        self.eta = eta
        self.tau = tau
        self.c = c
        self.dchi = dchi
        self.tau1 = tau1
        self.tau2 = tau2
        self.n_points_per_path = n_points_per_path

        if topo is None:
            topo = np.r_[0.0, 0.0, 0.0]
        self.topo = topo

        for src in self.survey.source_list:
            if np.any(src.location[..., 2] < self.topo[2]):
                raise ValueError("Source must be located above the topography")
            for rx in src.receiver_list:
                if rx.use_source_receiver_offset:
                    if np.any(src.location[2] + rx.locations[:, 2] < self.topo[2]):
                        raise ValueError(
                            "Receiver must be located above the topography"
                        )
                else:
                    if np.any(rx.locations[:, 2] < self.topo[2]):
                        raise ValueError(
                            "Receiver must be located above the topography"
                        )

        self.hankel_filter = hankel_filter
        self.fix_Jmatrix = fix_Jmatrix
        # Check input arguments. If self.hankel_filter is not a valid filter,
        # it will set it to the default (key_201_2009).
        ht, htarg = check_hankel(
            "dlf", {"dlf": self.hankel_filter, "pts_per_dec": 0}, 1
        )

        self._fhtfilt = htarg["dlf"]  # Store filter
        # self.hankel_pts_per_dec = htarg["pts_per_dec"]  # Store pts_per_dec
        if self.verbose:
            print(">> Use " + self.hankel_filter + " filter for Hankel Transform")

    @property
    def hankel_filter(self):
        """The hankely filter to use.

        Returns
        -------
        str
        """
        return self._hankel_filter

    @hankel_filter.setter
    def hankel_filter(self, value):
        self._hankel_filter = validate_string("hankel_filter", value)

    _hankel_pts_per_dec = 0  # Default: Standard DLF

    @property
    def fix_Jmatrix(self):
        """Whether to fix the sensitivity matrix.

        Returns
        -------
        bool
        """
        return self._fix_Jmatrix

    @fix_Jmatrix.setter
    def fix_Jmatrix(self, value):
        self._fix_Jmatrix = validate_type("fix_Jmatrix", value, bool)

    @property
    def topo(self):
        """Topography.

        Returns
        -------
        numpy.ndarray of float
        """
        return self._topo

    @topo.setter
    def topo(self, value):
        self._topo = validate_ndarray_with_shape("topo", value, shape=("*",))

    @property
    def n_layer(self):
        """number of layers"""
        return int(self.thicknesses.size + 1)

    @property
    def n_filter(self):
        """Length of filter"""
        return self._fhtfilt.base.size

    @property
    def depth(self):
        """layer depths"""
        if self.thicknesses is not None:
            return np.r_[0.0, -np.cumsum(self.thicknesses)]
        return None

    @property
    def n_points_per_path(self):
        """The number of integration points for each segment of line current sources.

        Returns
        -------
        int
        """
        return self._n_points_per_path

    @n_points_per_path.setter
    def n_points_per_path(self, val):
        self._n_points_per_path = validate_integer("n_points_per_path", val, min_val=1)

    def compute_complex_sigma(self, frequencies):
        r"""
        Computes the complex conductivity matrix using Pelton's Cole-Cole model:

        .. math ::

            \sigma (\omega ) = \sigma \Bigg [
                1 - \eta \Bigg ( \frac{1}{1 + (1-\eta ) (1 + i\omega \tau)^c} \Bigg )
            \Bigg ]

        :param numpy.array frequencies: np.array(N,) containing frequencies
        :rtype: numpy.ndarray: np.array(n_layer, n_frequency)
        :return: complex conductivity matrix

        """
        n_layer = self.n_layer
        n_frequency = len(frequencies)
        # n_filter = self.n_filter

        sigma = np.tile(self.sigma.reshape([-1, 1]), (1, n_frequency))

        # No IP effect
        if np.all(self.eta) == 0.0:
            return sigma

        # IP effect
        else:
            if np.isscalar(self.eta):
                eta = self.eta
                tau = self.tau
                c = self.c
            else:
                eta = np.tile(self.eta.reshape([-1, 1]), (1, n_frequency))
                tau = np.tile(self.tau.reshape([-1, 1]), (1, n_frequency))
                c = np.tile(self.c.reshape([-1, 1]), (1, n_frequency))

            w = np.tile(2 * np.pi * frequencies, (n_layer, 1))

            sigma_complex = np.empty(
                [n_layer, n_frequency], dtype=np.complex128, order="F"
            )
            sigma_complex[:, :] = sigma - sigma * eta / (
                1 + (1 - eta) * (1j * w * tau) ** c
            )

            return sigma_complex

    def compute_complex_mu(self, frequencies):
        r"""
        Computes the complex magnetic permeability matrix assuming a log-uniform
        distribution of time-relaxation constants:

        .. math::

            \chi (\omega ) = \chi + \Delta \chi \Bigg [
                1 - \Bigg ( \frac{1}{ln (\tau_2 / \tau_1 )} \Bigg )
                ln \Bigg ( \frac{1 + i\omega \tau_2}{1 + i\omega tau_1} ) \Bigg )
            \Bigg ]

        :param numpy.array frequencies: np.array(N,) containing frequencies
        :rtype: numpy.ndarray: np.array(n_layer, n_frequency)
        :return: complex magnetic susceptibility matrix
        """

        if np.isscalar(self.mu):
            mu = np.ones_like(self.sigma) * self.mu
        else:
            mu = self.mu

        n_layer = self.n_layer
        n_frequency = len(frequencies)
        # n_filter = self.n_filter

        mu = np.tile(mu.reshape([-1, 1]), (1, n_frequency))

        # No magnetic viscosity
        if np.all(self.dchi) == 0.0:
            return mu

        # Magnetic viscosity
        else:
            if np.isscalar(self.dchi):
                dchi = self.dchi * np.ones_like(self.mu)
                tau1 = self.tau1 * np.ones_like(self.mu)
                tau2 = self.tau2 * np.ones_like(self.mu)
            else:
                dchi = np.tile(self.dchi.reshape([-1, 1]), (1, n_frequency))
                tau1 = np.tile(self.tau1.reshape([-1, 1]), (1, n_frequency))
                tau2 = np.tile(self.tau2.reshape([-1, 1]), (1, n_frequency))

            w = np.tile(2 * np.pi * frequencies, (n_layer, 1))

            mu_complex = mu + mu_0 * dchi * (
                1
                - np.log((1 + 1j * w * tau2) / (1 + 1j * w * tau1))
                / np.log(tau2 / tau1)
            )

            return mu_complex

    def Jvec(self, m, v, f=None):
        Js = self.getJ(m, f=f)
        out = 0.0
        if self.hMap is not None:
            out = out + Js["dh"] @ (self.hDeriv @ v)
        if self.sigmaMap is not None:
            out = out + Js["ds"] @ (self.sigmaDeriv @ v)
        if self.muMap is not None:
            out = out + Js["dmu"] @ (self.muDeriv @ v)
        if self.thicknessesMap is not None:
            out = out + Js["dthick"] @ (self.thicknessesDeriv @ v)
        return out

    def Jtvec(self, m, v, f=None):
        Js = self.getJ(m, f=f)
        out = 0.0
        if self.hMap is not None:
            out = out + self.hDeriv.T @ (Js["dh"].T @ v)
        if self.sigmaMap is not None:
            out = out + self.sigmaDeriv.T @ (Js["ds"].T @ v)
        if self.muMap is not None:
            out = out + self.muDeriv.T @ (Js["dmu"].T @ v)
        if self.thicknessesMap is not None:
            out = out + self.thicknessesDeriv.T @ (Js["dthick"].T @ v)
        return out

    def _compute_hankel_coefficients(self):
        survey = self.survey
        C0s = []
        C1s = []
        lambs = []
        Is = []
        n_w_past = 0
        i_count = 0
        for src in survey.source_list:
            # doing the check for source type by checking its name
            # to avoid importing and checking "isinstance"
            class_name = type(src).__name__
            is_circular_loop = class_name == "CircularLoop"
            is_mag_dipole = class_name == "MagDipole"
            is_wire_loop = class_name == "LineCurrent"

            if is_circular_loop:
                if np.any(src.orientation[:-1] != 0.0):
                    raise ValueError("Can only simulate horizontal circular loops")
            if self.hMap is not None:
                h = 0  # source height above topo
            else:
                h = src.location[2] - self.topo[-1]

            if is_circular_loop or is_mag_dipole:
                src_x, src_y, src_z = src.orientation * src.moment / (4 * np.pi)
                # src.moment is pi * radius**2 * I for circular loop
            if is_wire_loop:
                x, w = roots_legendre(self.n_points_per_path)
                xy_src_path = src.location[:, :2]
                xyks = []
                thetas = []
                weights = []
                for i_path in range(src.n_segments):
                    dx = xy_src_path[i_path + 1, 0] - xy_src_path[i_path, 0]
                    dy = xy_src_path[i_path + 1, 1] - xy_src_path[i_path, 1]
                    dl = np.sqrt(dx**2 + dy**2)
                    theta = np.arctan2(dy, dx)
                    lk = np.c_[(x + 1) * dl / 2, np.zeros(self.n_points_per_path)]

                    R = np.array([[dx, -dy], [dy, dx]]) / dl
                    xyk = lk.dot(R.T) + xy_src_path[i_path, :]

                    xyks.append(xyk)
                    thetas.append(theta * np.ones(xyk.shape[0]))
                    weights.append(w * dl / 2)
                # store these for future evalution of integrals
                xyks = np.vstack(xyks)
                weights = np.hstack(weights) * src.current
                thetas = -np.hstack(thetas)

            for rx in src.receiver_list:
                #######
                # Hankel Transform coefficients
                ######
                rx_x, rx_y, rx_z = rx.orientation

                # Compute receiver height
                if rx.use_source_receiver_offset:
                    dxyz = rx.locations
                    z = h + rx.locations[:, 2]
                else:
                    dxyz = rx.locations - src.location
                    z = h + rx.locations[:, 2] - src.location[2]

                if is_wire_loop:
                    dxy = rx.locations[:, :2] - xyks
                    h = src.location.mean(axis=0)[2] - self.topo[-1]
                    z = h + rx.locations[:, 2] - src.location.mean(axis=0)[2]
                    offsets = np.linalg.norm(dxy, axis=-1)
                else:
                    offsets = np.linalg.norm(dxyz[:, :-1], axis=-1)

                if is_circular_loop:
                    if np.any(offsets != 0.0):
                        raise ValueError(
                            "Can only simulate central loop receivers with circular loop source"
                        )
                    offsets = src.radius * np.ones(rx.locations.shape[0])

                # computations for hankel transform...
                lambd, _ = get_dlf_points(
                    self._fhtfilt, offsets, self._hankel_pts_per_dec
                )
                # calculate the source-rx coefficients for the hankel transform
                C0 = 0.0
                C1 = 0.0
                if is_circular_loop:
                    # I * a/ 2 * (lambda **2 )/ (lambda)
                    C1 += src_z * rx_z * (2 / src.radius) * lambd
                    n_w = 1

                elif is_mag_dipole:
                    n_w = 1
                    if src_x != 0.0:
                        if rx_x != 0.0:
                            C0 += (
                                src_x
                                * rx_x
                                * (dxyz[:, 0] ** 2 / offsets**2)[:, None]
                                * lambd**2
                            )
                            C1 += (
                                src_x
                                * rx_x
                                * (1 / offsets - 2 * dxyz[:, 0] ** 2 / offsets**3)[
                                    :, None
                                ]
                                * lambd
                            )
                        if rx_y:
                            C0 += (
                                src_x
                                * rx_y
                                * (dxyz[:, 0] * dxyz[:, 1] / offsets**2)[:, None]
                                * lambd**2
                            )
                            C1 -= (
                                src_x
                                * rx_y
                                * (2 * dxyz[:, 0] * dxyz[:, 1] / offsets**3)[:, None]
                                * lambd
                            )
                        if rx_z != 0.0:
                            # C0 += 0.0
                            C1 -= (src_x * rx_z * dxyz[:, 0] / offsets)[
                                :, None
                            ] * lambd**2
                    if src_y != 0.0:
                        if rx_x != 0.0:
                            C0 += (
                                src_y
                                * rx_x
                                * rx_x
                                * (dxyz[:, 0] * dxyz[:, 1] / offsets**2)[:, None]
                                * lambd**2
                            )
                            C1 -= (
                                src_y
                                * rx_x
                                * (2 * dxyz[:, 0] * dxyz[:, 1] / offsets**3)[:, None]
                                * lambd
                            )
                        if rx_y != 0.0:
                            C0 += (
                                src_y
                                * rx_y
                                * (dxyz[:, 1] ** 2 / offsets**2)[:, None]
                                * lambd**2
                            )
                            C1 += (
                                src_y
                                * rx_y
                                * (1 / offsets - 2 * dxyz[:, 1] ** 2 / offsets**3)[
                                    :, None
                                ]
                                * lambd
                            )
                        if rx_z != 0.0:
                            # C0 += 0.0
                            C1 -= (src_y * rx_z * dxyz[:, 1] / offsets)[
                                :, None
                            ] * lambd**2
                    if src_z != 0.0:
                        if rx_x != 0.0:
                            # C0 += 0.0
                            C1 += (src_z * rx_x * dxyz[:, 0] / offsets)[
                                :, None
                            ] * lambd**2
                        if rx_y != 0.0:
                            # C0 += 0.0
                            C1 += (src_z * rx_y * dxyz[:, 1] / offsets)[
                                :, None
                            ] * lambd**2
                        if rx_z != 0.0:
                            C0 += src_z * rx_z * lambd**2
                elif is_wire_loop:
                    R = np.stack(
                        [
                            [np.cos(thetas), -np.sin(thetas)],
                            [np.sin(thetas), np.cos(thetas)],
                        ]
                    )
                    dxy_rot = np.einsum("...i,ji...", dxy, R)
                    C1 = (1 / (4 * np.pi) * (dxy_rot[:, 1] / offsets * weights))[
                        :, None
                    ] * lambd
                    # Assume
                    # 1) source_list only includes wire_loop sources
                    # 1) rx.locations.shape = (1,3)
                    n_w = weights.size
                else:
                    raise TypeError(
                        f"Unsupported source type of {type(src)}. Must be a CircularLoop or MagDipole"
                    )
                # divide by offsets to pre-do that part from the dft (1 less item to store)
                C0s.append(np.exp(-lambd * (z + h)[:, None]) * C0 / offsets[:, None])
                C1s.append(np.exp(-lambd * (z + h)[:, None]) * C1 / offsets[:, None])
                lambs.append(lambd)
                n_w_past += n_w
                Is.append(np.ones(n_w, dtype=int) * i_count)
                i_count += 1

        # Store these on the simulation for faster future executions
        self._lambs = np.vstack(lambs)
        self._unique_lambs, inv_lambs = np.unique(self._lambs, return_inverse=True)
        self._inv_lambs = inv_lambs.reshape(self._lambs.shape)
        self._C0s = np.vstack(C0s)
        self._C1s = np.vstack(C1s)
        Is = np.hstack(Is)
        n_row = Is.size
        n_col = Is.max() + 1
        Js = np.arange(n_row)
        data = np.ones(n_row, dtype=int)
        self._W = sp.coo_matrix((data, (Is, Js)), shape=(n_col, n_row))
        self._W = self._W.tocsr()

    @property
    def deleteTheseOnModelUpdate(self):
        toDelete = super().deleteTheseOnModelUpdate
        if self.fix_Jmatrix is False:
            toDelete += ["_J", "_jtjdiag"]
        return toDelete

    def depth_of_investigation_christiansen_2012(self, std, thres_hold=0.8):
        pred = self.survey._pred.copy()
        delta_d = std * np.log(abs(self.survey.dobs))
        J = self.getJ(self.model)
        J_sum = abs(utils.sdiag(1 / delta_d / pred) * J).sum(axis=0)
        S = np.cumsum(J_sum[::-1])[::-1]
        active = S - thres_hold > 0.0
        doi = abs(self.depth[active]).max()
        return doi, active

    def get_threshold(self, uncert):
        _, active = self.depth_of_investigation(uncert)
        JtJdiag = self.get_JtJdiag(uncert)
        delta = JtJdiag[active].min()
        return delta

    def getJtJdiag(self, m, W=None, f=None):
        if getattr(self, "_jtjdiag", None) is None:
            Js = self.getJ(m, f=f)
            if W is None:
                W = np.ones(self.survey.nD)
            else:
                W = W.diagonal() ** 2
            out = 0.0
            if self.hMap is not None:
                J = Js["dh"] @ self.hDeriv
                out = out + np.einsum("i,ij,ij->j", W, J, J)
            if self.sigmaMap is not None:
                J = Js["ds"] @ self.sigmaDeriv
                out = out + np.einsum("i,ij,ij->j", W, J, J)
            if self.muMap is not None:
                J = Js["dmu"] @ self.muDeriv
                out = out + np.einsum("i,ij,ij->j", W, J, J)
            if self.thicknessesMap is not None:
                J = Js["dthick"] @ self.thicknessesDeriv
                out = out + np.einsum("i,ij,ij->j", W, J, J)
            self._jtjdiag = out
        return self._jtjdiag



"""
==============================================================================
Frequency-Domain EM (:mod:`SimPEG.electromagnetics.frequency_domain`)
==============================================================================
.. currentmodule:: SimPEG.electromagnetics.frequency_domain

About ``frequency_domain``

Simulations
===========
.. autosummary::
  :toctree: generated/

  Simulation1DLayered
  Simulation3DElectricField
  Simulation3DMagneticFluxDensity
  Simulation3DCurrentDensity
  Simulation3DMagneticField


Receivers
=========
.. autosummary::
  :toctree: generated/

  receivers.PointElectricField
  receivers.PointMagneticFluxDensity
  receivers.PointMagneticFluxDensitySecondary
  receivers.PointMagneticField
  receivers.PointCurrentDensity

Sources
=======
.. autosummary::
  :toctree: generated/

  sources.RawVec_e
  sources.RawVec_m
  sources.RawVec
  sources.MagDipole
  sources.MagDipole_Bfield
  sources.CircularLoop
  sources.PrimSecSigma
  sources.PrimSecMappedSigma
  sources.LineCurrent

Surveys
=======
.. autosummary::
  :toctree: generated/

  survey.Survey

Fields
======
.. autosummary::
  :toctree: generated/

  Fields3DElectricField
  Fields3DMagneticFluxDensity
  Fields3DCurrentDensity
  Fields3DMagneticField

Base Classes
============
.. autosummary::
  :toctree: generated/

  survey.Survey
  sources.BaseFDEMSrc
  receivers.BaseRx
  simulation.BaseFDEMSimulation
  fields.FieldsFDEM

"""
from .survey import Survey
from . import sources
from . import receivers
from .simulation import (
    Simulation3DElectricField,
    Simulation3DMagneticFluxDensity,
    Simulation3DCurrentDensity,
    Simulation3DMagneticField,
)
from .simulation_1d import Simulation1DLayered
from .fields import (
    Fields3DElectricField,
    Fields3DMagneticFluxDensity,
    Fields3DCurrentDensity,
    Fields3DMagneticField,
)

from . import sources as Src
from . import receivers as Rx



import numpy as np
from dask.distributed import get_client
from multiprocessing import cpu_count


def compute_chunk_sizes(M, N, target_chunk_size):
    """
    Compute row and collumn chunk sizes for a matrix of shape MxN,
    such that the chunks are below a certain threshold target_chunk_size (in Mb)
    """
    nChunks_col = 1
    nChunks_row = 1
    rowChunk = int(np.ceil(M / nChunks_row))
    colChunk = int(np.ceil(N / nChunks_col))
    chunk_size = rowChunk * colChunk * 8 * 1e-6

    # Add more chunks until memory falls below target
    while chunk_size >= target_chunk_size:
        if rowChunk > colChunk:
            nChunks_row += 1
        else:
            nChunks_col += 1

        rowChunk = int(np.ceil(M / nChunks_row))
        colChunk = int(np.ceil(N / nChunks_col))
        chunk_size = rowChunk * colChunk * 8 * 1e-6  # in Mb
    return rowChunk, colChunk


def compute(self, job):
    """
    Compute dask job for either dask array or client.
    """
    if isinstance(job, np.ndarray):
        return job
    try:
        client = get_client()
        return client.compute(job, workers=self.workers)
    except ValueError:
        return job.compute()


def get_parallel_blocks(source_list: list, m_size: int, max_chunk_size: int) -> list:
    """
    Get the blocks of sources and receivers to be computed in parallel.

    Stored as a list of tuples for
    (source, receiver, block index) and array of indices
    for the rows of the sensitivity matrix.
    """
    data_block_size = np.ceil(max_chunk_size / (m_size * 8.0 * 1e-6))
    row_count = 0
    row_index = 0
    block_count = 0
    blocks = [[]]

    for s_id, src in enumerate(source_list):
        for r_id, rx in enumerate(src.receiver_list):
            indices = np.arange(rx.nD).astype(int)
            chunks = np.array_split(
                indices, int(np.ceil(len(indices) / data_block_size))
            )

            for ind, chunk in enumerate(chunks):
                chunk_size = len(chunk)

                # Condition to start a new block
                if (row_count + chunk_size) > (data_block_size * cpu_count()):
                    row_count = 0
                    block_count += 1
                    blocks.append([])

                blocks[block_count].append(
                    (
                        (s_id, r_id, ind),
                        (
                            chunk,
                            np.arange(row_index, row_index + chunk_size).astype(int),
                            rx.locations.shape[0],
                        ),
                    )
                )
                row_index += chunk_size
                row_count += chunk_size

    # Re-split over cpu_count if too few blocks
    if len(blocks) < cpu_count():
        flatten_blocks = []
        for block in blocks:
            flatten_blocks += block

        chunks = np.array_split(np.arange(len(flatten_blocks)), cpu_count())
        return [[flatten_blocks[i] for i in chunk] for chunk in chunks]
    return blocks



from ..resistivity import Survey
from ..resistivity import receivers, sources  # noqa: F401


def from_dc_to_ip_survey(dc_survey, dim="2.5D"):
    """Create IP survey from DC survey geometry

    Parameters
    ----------
    dc_survey : SimPEG.electromagnetics.static.resistivity.survey.Survey
        DC survey object
    dim : {'2.5D', '1D', '3D'}
        Dimension of the surface.

    Returns
    -------
    SimPEG.electromagnetics.static.induced_polarization.survey.Survey
        An IP survey object
    """
    source_list = dc_survey.source_list
    ip_survey = Survey(source_list)

    return ip_survey
